"""
The measurement module provides a context manager for registering parameters
to measure and storing results. The user is expected to mainly interact with it
using the :class:`.Measurement` class.
"""

import collections
import io
import logging
import traceback as tb_module
import warnings
from copy import deepcopy
from inspect import signature
from numbers import Number
from time import perf_counter
from types import TracebackType
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    List,
    Mapping,
    MutableMapping,
    MutableSequence,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
)

import numpy as np

import qcodes as qc
import qcodes.utils.validators as vals
from qcodes.dataset.data_set import (
    VALUE,
    DataSet,
    load_by_guid,
    res_type,
    setpoints_type,
    values_type,
)
from qcodes.dataset.data_set_protocol import DataSetProtocol
from qcodes.dataset.descriptions.dependencies import (
    DependencyError,
    InferenceError,
    InterDependencies_,
)
from qcodes.dataset.descriptions.param_spec import ParamSpec, ParamSpecBase
from qcodes.dataset.descriptions.rundescriber import RunDescriber
from qcodes.dataset.descriptions.versioning.rundescribertypes import Shapes
from qcodes.dataset.experiment_container import Experiment
from qcodes.dataset.export_config import get_data_export_automatic
from qcodes.instrument.delegate.grouped_parameter import GroupedParameter
from qcodes.instrument.parameter import (
    ArrayParameter,
    MultiParameter,
    Parameter,
    ParameterWithSetpoints,
    _BaseParameter,
    expand_setpoints_helper,
)
from qcodes.station import Station
from qcodes.utils.delaykeyboardinterrupt import DelayedKeyboardInterrupt

if TYPE_CHECKING:
    from qcodes.dataset.sqlite.connection import ConnectionPlus

log = logging.getLogger(__name__)


ActionType = Tuple[Callable[..., Any], Sequence[Any]]
SubscriberType = Tuple[Callable[..., Any],
                       Union[MutableSequence[Any],
                             MutableMapping[Any, Any]]]


class ParameterTypeError(Exception):
    pass


class DataSaver:
    """
    The class used by the :class:`Runner` context manager to handle the
    datasaving to the database.
    """

    default_callback: Optional[Dict[Any, Any]] = None

    def __init__(
        self,
        dataset: DataSetProtocol,
        write_period: float,
        interdeps: InterDependencies_,
    ) -> None:
        self._dataset = dataset
        if (
            DataSaver.default_callback is not None
            and "run_tables_subscription_callback" in DataSaver.default_callback
        ):
            callback = DataSaver.default_callback["run_tables_subscription_callback"]
            min_wait = DataSaver.default_callback["run_tables_subscription_min_wait"]
            min_count = DataSaver.default_callback["run_tables_subscription_min_count"]
            snapshot = dataset.metadata["snapshot"]
            if isinstance(self._dataset, DataSet):
                self._dataset.subscribe(
                    callback,
                    min_wait=min_wait,
                    min_count=min_count,
                    state={},
                    callback_kwargs={
                        "run_id": self._dataset.run_id,
                        "snapshot": snapshot,
                    },
                )
        if isinstance(self._dataset, DataSet):
            default_subscribers = qc.config.subscription.default_subscribers
            for subscriber in default_subscribers:
                self._dataset.subscribe_from_config(subscriber)

        self._interdeps = interdeps
        self.write_period = float(write_period)
        # self._results will be filled by add_result
        self._results: List[Dict[str, VALUE]] = []
        self._last_save_time = perf_counter()
        self._known_dependencies: Dict[str, List[str]] = {}
        self.parent_datasets: List[DataSet] = []

        for link in self._dataset.parent_dataset_links:
            self.parent_datasets.append(load_by_guid(link.tail))

    def add_result(self, *res_tuple: res_type) -> None:
        """
        Add a result to the measurement results. Represents a measurement
        point in the space of measurement parameters, e.g. in an experiment
        varying two voltages and measuring two currents, a measurement point
        is four dimensional (v1, v2, c1, c2). The corresponding call
        to this function would be

            >>> datasaver.add_result((v1, 0.1), (v2, 0.2), (c1, 5), (c2, -2.1))

        For better performance, this function does not immediately write to
        the database, but keeps the results in memory. Writing happens every
        ``write_period`` seconds and during the ``__exit__`` method
        of this class.

        Args:
            res_tuple: A tuple with the first element being the parameter name
                and the second element is the corresponding value(s) at this
                measurement point. The function takes as many tuples as there
                are results.

        Raises:
            ValueError: If a parameter name is not registered in the parent
                Measurement object.
            ValueError: If the shapes of parameters do not match, i.e. if a
                parameter gets values of a different shape than its setpoints
                (the exception being that setpoints can always be scalar)
            ValueError: If multiple results are given for the same parameter.
            ParameterTypeError: If a parameter is given a value not matching
                its type.
        """

        # we iterate through the input twice. First we find any array and
        # multiparameters that need to be unbundled and collect the names
        # of all parameters. This also allows users to call
        # add_result with the arguments in any particular order, i.e. NOT
        # enforcing that setpoints come before dependent variables.
        results_dict: Dict[ParamSpecBase, np.ndarray] = {}

        parameter_names = tuple(partial_result[0].full_name
                                if isinstance(partial_result[0], _BaseParameter) else partial_result[0]
                                for partial_result in res_tuple)
        if len(set(parameter_names)) != len(parameter_names):
            non_unique = [
                item
                for item, count in collections.Counter(parameter_names).items()
                if count > 1
            ]
            raise ValueError(
                f"Not all parameter names are unique. "
                f"Got multiple values for {non_unique}"
            )

        for partial_result in res_tuple:
            parameter = partial_result[0]
            data = partial_result[1]

            if (isinstance(parameter, _BaseParameter) and
                    isinstance(parameter.vals, vals.Arrays)):
                if not isinstance(data, np.ndarray):
                    raise TypeError(
                        f"Expected data for Parameter with Array validator "
                        f"to be a numpy array but got: {type(data)}")

                if (parameter.vals.shape is not None
                        and data.shape != parameter.vals.shape):
                    raise TypeError(
                        f"Expected data with shape {parameter.vals.shape}, "
                        f"but got {data.shape} for parameter: {parameter.full_name}"
                    )

            if isinstance(parameter, ArrayParameter):
                results_dict.update(
                    self._unpack_arrayparameter(partial_result))
            elif isinstance(parameter, MultiParameter):
                results_dict.update(
                    self._unpack_multiparameter(partial_result))
            elif isinstance(parameter, ParameterWithSetpoints):
                results_dict.update(
                    self._conditionally_expand_parameter_with_setpoints(
                        data, parameter, parameter_names, partial_result
                    )
                )
            else:
                results_dict.update(
                    self._unpack_partial_result(partial_result)
                )

        self._validate_result_deps(results_dict)
        self._validate_result_shapes(results_dict)
        self._validate_result_types(results_dict)

        self.dataset._enqueue_results(results_dict)

        if perf_counter() - self._last_save_time > self.write_period:
            self.flush_data_to_database()
            self._last_save_time = perf_counter()

    def _conditionally_expand_parameter_with_setpoints(
            self, data: values_type, parameter: ParameterWithSetpoints,
            parameter_names: Sequence[str], partial_result: res_type
    ) -> Dict[ParamSpecBase, np.ndarray]:
        local_results = {}
        setpoint_names = tuple(setpoint.full_name for setpoint in parameter.setpoints)
        expanded = tuple(setpoint_name in parameter_names for setpoint_name in setpoint_names)
        if all(expanded):
            local_results.update(
                self._unpack_partial_result(partial_result))
        elif any(expanded):
            raise ValueError(f"Some of the setpoints of {parameter.full_name} "
                             "were explicitly given but others were not. "
                             "Either supply all of them or none of them.")
        else:
            expanded_partial_result = expand_setpoints_helper(parameter, data)
            for res in expanded_partial_result:
                local_results.update(
                    self._unpack_partial_result(res)
                )
        return local_results

    def _unpack_partial_result(
            self,
            partial_result: res_type) -> Dict[ParamSpecBase, np.ndarray]:
        """
        Unpack a partial result (not containing :class:`ArrayParameters` or
        class:`MultiParameters`) into a standard results dict form and return
        that dict
        """
        param, values = partial_result
        try:
            parameter = self._interdeps._id_to_paramspec[str(param)]
        except KeyError:
            raise ValueError('Can not add result for parameter '
                             f'{param}, no such parameter registered '
                             'with this measurement.')
        return {parameter: np.array(values)}

    def _unpack_arrayparameter(
        self, partial_result: res_type) -> Dict[ParamSpecBase, np.ndarray]:
        """
        Unpack a partial result containing an :class:`Arrayparameter` into a
        standard results dict form and return that dict
        """
        array_param, values_array = partial_result
        array_param = cast(ArrayParameter, array_param)

        if array_param.setpoints is None:
            raise RuntimeError(f"{array_param.full_name} is an "
                               f"{type(array_param)} "
                               f"without setpoints. Cannot handle this.")
        try:
            main_parameter = self._interdeps._id_to_paramspec[str(array_param)]
        except KeyError:
            raise ValueError('Can not add result for parameter '
                             f'{array_param}, no such parameter registered '
                             'with this measurement.')

        res_dict = {main_parameter: np.array(values_array)}

        sp_names = array_param.setpoint_full_names
        fallback_sp_name = f"{array_param.full_name}_setpoint"

        res_dict.update(
            self._unpack_setpoints_from_parameter(
                array_param, array_param.setpoints,
                sp_names, fallback_sp_name))

        return res_dict

    def _unpack_multiparameter(
            self, partial_result: res_type) -> Dict[ParamSpecBase, np.ndarray]:
        """
        Unpack the `subarrays` and `setpoints` from a :class:`MultiParameter`
        and into a standard results dict form and return that dict

        """

        parameter, data = partial_result
        parameter = cast(MultiParameter, parameter)

        result_dict = {}

        if parameter.setpoints is None:
            raise RuntimeError(f"{parameter.full_name} is an "
                               f"{type(parameter)} "
                               f"without setpoints. Cannot handle this.")
        for i in range(len(parameter.shapes)):
            # if this loop runs, then 'data' is a Sequence
            data = cast(Sequence[Union[str, int, float, Any]], data)

            shape = parameter.shapes[i]

            try:
                paramspec = self._interdeps._id_to_paramspec[parameter.full_names[i]]
            except KeyError:
                raise ValueError('Can not add result for parameter '
                                 f'{parameter.names[i]}, '
                                 'no such parameter registered '
                                 'with this measurement.')

            result_dict.update({paramspec: np.array(data[i])})
            if shape != ():
                # array parameter like part of the multiparameter
                # need to find setpoints too
                fallback_sp_name = f'{parameter.full_names[i]}_setpoint'

                sp_names: Optional[Sequence[str]]
                if (parameter.setpoint_full_names is not None
                        and parameter.setpoint_full_names[i] is not None):
                    sp_names = parameter.setpoint_full_names[i]
                else:
                    sp_names = None

                result_dict.update(
                    self._unpack_setpoints_from_parameter(
                        parameter,
                        parameter.setpoints[i],
                        sp_names,
                        fallback_sp_name))

        return result_dict

    def _unpack_setpoints_from_parameter(
        self, parameter: _BaseParameter, setpoints: Sequence[Any],
        sp_names: Optional[Sequence[str]], fallback_sp_name: str
            ) -> Dict[ParamSpecBase, np.ndarray]:
        """
        Unpack the `setpoints` and their values from a
        :class:`ArrayParameter` or :class:`MultiParameter`
        into a standard results dict form and return that dict
        """
        setpoint_axes = []
        setpoint_parameters: List[ParamSpecBase] = []

        for i, sps in enumerate(setpoints):
            if sp_names is not None:
                spname = sp_names[i]
            else:
                spname = f'{fallback_sp_name}_{i}'

            try:
                setpoint_parameter = self._interdeps[spname]
            except KeyError:
                raise RuntimeError('No setpoints registered for '
                                   f'{type(parameter)} {parameter.full_name}!')
            sps = np.array(sps)
            while sps.ndim > 1:
                # The outermost setpoint axis or an nD param is nD
                # but the innermost is 1D. In all cases we just need
                # the axis along one dim, the innermost one.
                sps = sps[0]

            setpoint_parameters.append(setpoint_parameter)
            setpoint_axes.append(sps)

        output_grids = np.meshgrid(*setpoint_axes, indexing='ij')
        result_dict = {}
        for grid, param in zip(output_grids, setpoint_parameters):
            result_dict.update({param: grid})

        return result_dict

    def _validate_result_deps(
            self, results_dict: Mapping[ParamSpecBase, values_type]) -> None:
        """
        Validate that the dependencies of the ``results_dict`` are met,
        meaning that (some) values for all required setpoints and inferences
        are present
        """
        try:
            self._interdeps.validate_subset(list(results_dict.keys()))
        except (DependencyError, InferenceError) as err:
            raise ValueError('Can not add result, some required parameters '
                             'are missing.') from err

    def _validate_result_shapes(
            self, results_dict: Mapping[ParamSpecBase, values_type]) -> None:
        """
        Validate that all sizes of the ``results_dict`` are consistent.
        This means that array-values of parameters and their setpoints are
        of the same size, whereas parameters with no setpoint relation to
        each other can have different sizes.
        """
        toplevel_params = (set(self._interdeps.dependencies)
                           .intersection(set(results_dict)))
        for toplevel_param in toplevel_params:
            required_shape = np.shape(np.array(results_dict[toplevel_param]))
            for setpoint in self._interdeps.dependencies[toplevel_param]:
                # a setpoint is allowed to be a scalar; shape is then ()
                setpoint_shape = np.shape(np.array(results_dict[setpoint]))
                if setpoint_shape not in [(), required_shape]:
                    raise ValueError(f'Incompatible shapes. Parameter '
                                     f"{toplevel_param.name} has shape "
                                     f"{required_shape}, but its setpoint "
                                     f"{setpoint.name} has shape "
                                     f"{setpoint_shape}.")

    @staticmethod
    def _validate_result_types(
            results_dict: Mapping[ParamSpecBase, np.ndarray]) -> None:
        """
        Validate the type of the results
        """

        allowed_kinds = {'numeric': 'iuf', 'text': 'SU', 'array': 'iufcSUmM',
                         'complex': 'c'}

        for ps, vals in results_dict.items():
                if vals.dtype.kind not in allowed_kinds[ps.type]:
                    raise ValueError(f'Parameter {ps.name} is of type '
                                     f'"{ps.type}", but got a result of '
                                     f'type {vals.dtype} ({vals}).')

    def flush_data_to_database(self, block: bool = False) -> None:
        """
        Write the in-memory results to the database.

        Args:
            block: If writing using a background thread block until the
                background thread has written all data to disc. The
                argument has no effect if not using a background thread.

        """
        self.dataset._flush_data_to_database(block=block)

    def export_data(self) -> None:
        """Export data at end of measurement as per export_type
        specification in "dataset" section of qcodes config
        """
        self.dataset.export()

    @property
    def run_id(self) -> int:
        return self._dataset.run_id

    @property
    def points_written(self) -> int:
        return self._dataset.number_of_results

    @property
    def dataset(self) -> DataSetProtocol:
        return self._dataset


class Runner:
    """
    Context manager for the measurement.

    Lives inside a :class:`Measurement` and should never be instantiated
    outside a Measurement.

    This context manager handles all the dirty business of writing data
    to the database. Additionally, it may perform experiment bootstrapping
    and clean-up after a measurement.
    """

    def __init__(
        self,
        enteractions: Sequence[ActionType],
        exitactions: Sequence[ActionType],
        experiment: Optional[Experiment] = None,
        station: Optional[Station] = None,
        write_period: Optional[float] = None,
        interdeps: InterDependencies_ = InterDependencies_(),
        name: str = "",
        subscribers: Optional[Sequence[SubscriberType]] = None,
        parent_datasets: Sequence[Mapping[Any, Any]] = (),
        extra_log_info: str = "",
        write_in_background: bool = False,
        shapes: Optional[Shapes] = None,
        in_memory_cache: bool = True,
        dataset_class: Type[DataSetProtocol] = DataSet,
    ) -> None:

        self._dataset_class = dataset_class
        self.write_period = self._calculate_write_period(write_in_background,
                                                         write_period)

        self.enteractions = enteractions
        self.exitactions = exitactions
        self.subscribers: Sequence[SubscriberType]
        if subscribers is None:
            self.subscribers = []
        else:
            self.subscribers = subscribers
        self.experiment = experiment
        self.station = station
        self._interdependencies = interdeps
        self._shapes: Shapes = shapes
        self.name = name if name else 'results'
        self._parent_datasets = parent_datasets
        self._extra_log_info = extra_log_info
        self._write_in_background = write_in_background
        self._in_memory_cache = in_memory_cache
        self.ds: DataSetProtocol

    @staticmethod
    def _calculate_write_period(
            write_in_background: bool,
            write_period: Optional[float]
    ) -> float:
        write_period_changed_from_default = (
                write_period is not None and
                write_period != qc.config.defaults.dataset.write_period
        )
        if write_in_background and write_period_changed_from_default:
            warnings.warn(f"The specified write period of {write_period} s "
                          "will be ignored, since write_in_background==True")
        if write_in_background:
            return 0.0
        if write_period is None:
            write_period = qc.config.dataset.write_period
        return float(write_period)

    def __enter__(self) -> DataSaver:
        # TODO: should user actions really precede the dataset?
        # first do whatever bootstrapping the user specified

        for func, args in self.enteractions:
            func(*args)

        dataset_class: Type[DataSetProtocol]

        # next set up the "datasaver"
        if self.experiment is not None:
            exp_id: Optional[int] = self.experiment.exp_id
            conn: Optional["ConnectionPlus"] = self.experiment.conn
        else:
            exp_id = None
            conn = None

        if self._dataset_class is DataSet:
            dataset_class = cast(Type[DataSet], self._dataset_class)
            self.ds = dataset_class(
                name=self.name,
                exp_id=exp_id,
                conn=conn,
                in_memory_cache=self._in_memory_cache,
            )
        else:
            raise RuntimeError("Does not support any other dataset classes")
        # .. and give the dataset a snapshot as metadata
        if self.station is None:
            station = qc.Station.default
        else:
            station = self.station

        if station is not None:
            snapshot = station.snapshot()
        else:
            snapshot = {}

        self.ds.prepare(
            snapshot=snapshot,
            interdeps=self._interdependencies,
            write_in_background=self._write_in_background,
            shapes=self._shapes,
            parent_datasets=self._parent_datasets,
        )

        # register all subscribers
        if isinstance(self.ds, DataSet):
            for (callble, state) in self.subscribers:
                # We register with minimal waiting time.
                # That should make all subscribers be called when data is flushed
                # to the database
                log.debug(f"Subscribing callable {callble} with state {state}")
                self.ds.subscribe(callble, min_wait=0, min_count=1, state=state)

        print(
            f"Starting experimental run with id: {self.ds.captured_run_id}."
            f" {self._extra_log_info}"
        )
        log.info(
            f"Starting measurement with guid: {self.ds.guid}, "
            f'sample_name: "{self.ds.sample_name}", '
            f'exp_name: "{self.ds.exp_name}", '
            f'ds_name: "{self.ds.name}". '
            f"{self._extra_log_info}"
        )
        log.info(f"Using background writing: {self._write_in_background}")

        self.datasaver = DataSaver(
                            dataset=self.ds,
                            write_period=self.write_period,
                            interdeps=self._interdependencies)

        return self.datasaver

    def __exit__(self,
                 exception_type: Optional[Type[BaseException]],
                 exception_value: Optional[BaseException],
                 traceback: Optional[TracebackType]
                 ) -> None:
        with DelayedKeyboardInterrupt():
            self.datasaver.flush_data_to_database(block=True)

            # perform the "teardown" events
            for func, args in self.exitactions:
                func(*args)

            if exception_type:
                # if an exception happened during the measurement,
                # log the exception
                stream = io.StringIO()
                tb_module.print_exception(exception_type,
                                          exception_value,
                                          traceback,
                                          file=stream)
                exception_string = stream.getvalue()
                log.warning('An exception occured in measurement with guid: '
                            f'{self.ds.guid};\nTraceback:\n{exception_string}')
                self.ds.add_metadata("measurement_exception", exception_string)

            # and finally mark the dataset as closed, thus
            # finishing the measurement
            # Note that the completion of a dataset entails waiting for the
            # write thread to terminate (iff the write thread has been started)
            self.ds.mark_completed()
            if get_data_export_automatic():
                self.datasaver.export_data()
            log.info(f'Finished measurement with guid: {self.ds.guid}. '
                     f'{self._extra_log_info}')
            if isinstance(self.ds, DataSet):
                self.ds.unsubscribe_all()


T = TypeVar('T', bound='Measurement')


class Measurement:
    """
    Measurement procedure container. Note that multiple measurement
    instances cannot be nested.

    Args:
        exp: Specify the experiment to use. If not given
            the default one is used. The default experiment
            is the latest one created.
        station: The QCoDeS station to snapshot. If not given, the
            default one is used.
        name: Name of the measurement. This will be passed down to the dataset
            produced by the measurement. If not given, a default value of
            'results' is used for the dataset.
    """

    def __init__(self, exp: Optional[Experiment] = None,
                 station: Optional[qc.Station] = None,
                 name: str = '') -> None:
        self.exitactions: List[ActionType] = []
        self.enteractions: List[ActionType] = []
        self.subscribers: List[SubscriberType] = []

        self.experiment = exp
        self.station = station
        self.name = name
        self.write_period: float = qc.config.dataset.write_period
        self._interdeps = InterDependencies_()
        self._shapes: Shapes = None
        self._parent_datasets: List[Dict[str, str]] = []
        self._extra_log_info: str = ''

    @property
    def parameters(self) -> Dict[str, ParamSpecBase]:
        return deepcopy(self._interdeps._id_to_paramspec)

    @property
    def write_period(self) -> float:
        return self._write_period

    @write_period.setter
    def write_period(self, wp: float) -> None:
        if not isinstance(wp, Number):
            raise ValueError('The write period must be a number (of seconds).')
        wp_float = float(wp)
        if wp_float < 1e-3:
            raise ValueError('The write period must be at least 1 ms.')
        self._write_period = wp_float

    def _paramspecbase_from_strings(
            self, name: str, setpoints: Optional[Sequence[str]] = None,
            basis: Optional[Sequence[str]] = None
            ) -> Tuple[Tuple[ParamSpecBase, ...], Tuple[ParamSpecBase, ...]]:
        """
        Helper function to look up and get ParamSpecBases and to give a nice
        error message if the user tries to register a parameter with reference
        (setpoints, basis) to a parameter not registered with this measurement

        Called by _register_parameter only.

        Args:
            name: Name of the parameter to register
            setpoints: name(s) of the setpoint parameter(s)
            basis: name(s) of the parameter(s) that this parameter is
                inferred from
        """

        idps = self._interdeps

        # now handle setpoints
        depends_on = []
        if setpoints:
            for sp in setpoints:
                try:
                    sp_psb = idps._id_to_paramspec[sp]
                    depends_on.append(sp_psb)
                except KeyError:
                    raise ValueError(f'Unknown setpoint: {sp}.'
                                     ' Please register that parameter first.')

        # now handle inferred parameters
        inf_from = []
        if basis:
            for inff in basis:
                try:
                    inff_psb = idps._id_to_paramspec[inff]
                    inf_from.append(inff_psb)
                except KeyError:
                    raise ValueError(f'Unknown basis parameter: {inff}.'
                                     ' Please register that parameter first.')

        return tuple(depends_on), tuple(inf_from)

    def register_parent(
            self: T, parent: DataSet, link_type: str,
            description: str = "") -> T:
        """
        Register a parent for the outcome of this measurement

        Args:
            parent: The parent dataset
            link_type: A name for the type of parent-child link
            description: A free-text description of the relationship
        """
        # we save the information in a way that is very compatible with the
        # Link object we will eventually make out of this information. We
        # cannot create a Link object just yet, because the DataSet of this
        # Measurement has not been given a GUID yet
        parent_dict = {'tail': parent.guid, 'edge_type': link_type,
                       'description': description}
        self._parent_datasets.append(parent_dict)

        return self

    def register_parameter(
            self: T, parameter: _BaseParameter,
            setpoints: Optional[setpoints_type] = None,
            basis: Optional[setpoints_type] = None,
            paramtype: Optional[str] = None) -> T:
        """
        Add QCoDeS Parameter to the dataset produced by running this
        measurement.

        Args:
            parameter: The parameter to add
            setpoints: The Parameter representing the setpoints for this
                parameter. If this parameter is a setpoint,
                it should be left blank
            basis: The parameters that this parameter is inferred from. If
                this parameter is not inferred from any other parameters,
                this should be left blank.
            paramtype: Type of the parameter, i.e. the SQL storage class,
                If None the paramtype will be inferred from the parameter type
                and the validator of the supplied parameter.
        """
        if not isinstance(parameter, _BaseParameter):
            raise ValueError('Can not register object of type {}. Can only '
                             'register a QCoDeS Parameter.'
                             ''.format(type(parameter)))

        paramtype = self._infer_paramtype(parameter, paramtype)
        # default to numeric
        if paramtype is None:
            paramtype = 'numeric'

        # now the parameter type must be valid
        if paramtype not in ParamSpec.allowed_types:
            raise RuntimeError("Trying to register a parameter with type "
                               f"{paramtype}. However, only "
                               f"{ParamSpec.allowed_types} are supported.")

        if isinstance(parameter, ArrayParameter):
            self._register_arrayparameter(parameter,
                                          setpoints,
                                          basis,
                                          paramtype)
        elif isinstance(parameter, ParameterWithSetpoints):
            self._register_parameter_with_setpoints(parameter,
                                                    setpoints,
                                                    basis,
                                                    paramtype)
        elif isinstance(parameter, MultiParameter):
            self._register_multiparameter(parameter,
                                          setpoints,
                                          basis,
                                          paramtype,
                                          )
        elif isinstance(parameter, Parameter):
            self._register_parameter(parameter.full_name,
                                     parameter.label,
                                     parameter.unit,
                                     setpoints,
                                     basis, paramtype)
        elif isinstance(parameter, GroupedParameter):
            self._register_parameter(parameter.full_name,
                                     parameter.label,
                                     parameter.unit,
                                     setpoints,
                                     basis, paramtype)
        else:
            raise RuntimeError("Does not know how to register a parameter"
                               f"of type {type(parameter)}")

        return self

    @staticmethod
    def _infer_paramtype(parameter: _BaseParameter,
                         paramtype: Optional[str]) -> Optional[str]:
        """
        Infer the best parameter type to store the parameter supplied.

        Args:
            parameter: The parameter to to infer the type for
            paramtype: The initial supplied parameter type or None

        Returns:
            The inferred parameter type. If a not None parameter type is
            supplied this will be preferred over any inferred type.
            Returns None if a parameter type could not be inferred
        """
        if paramtype is not None:
            return paramtype

        if isinstance(parameter.vals, vals.Arrays):
            paramtype = 'array'
        elif isinstance(parameter, ArrayParameter):
            paramtype = 'array'
        elif isinstance(parameter.vals, vals.Strings):
            paramtype = 'text'
        elif isinstance(parameter.vals, vals.ComplexNumbers):
            paramtype = 'complex'
        # TODO should we try to figure out if parts of a multiparameter are
        # arrays or something else?
        return paramtype

    def _register_parameter(self: T, name: str,
                            label: Optional[str],
                            unit: Optional[str],
                            setpoints: Optional[setpoints_type],
                            basis: Optional[setpoints_type],
                            paramtype: str) -> T:
        """
        Update the interdependencies object with a new group
        """

        parameter: Optional[ParamSpecBase]

        try:
            parameter = self._interdeps[name]
        except KeyError:
            parameter = None

        paramspec = ParamSpecBase(name=name,
                                  paramtype=paramtype,
                                  label=label,
                                  unit=unit)

        # We want to allow the registration of the exact same parameter twice,
        # the reason being that e.g. two ArrayParameters could share the same
        # setpoint parameter, which would then be registered along with each
        # dependent (array)parameter

        if parameter is not None and parameter != paramspec:
            raise ValueError("Parameter already registered "
                             "in this Measurement.")

        if setpoints is not None:
            sp_strings = [str(sp) for sp in setpoints]
        else:
            sp_strings = []

        if basis is not None:
            bs_strings = [str(bs) for bs in basis]
        else:
            bs_strings = []

        # get the ParamSpecBases
        depends_on, inf_from = self._paramspecbase_from_strings(name,
                                                                sp_strings,
                                                                bs_strings)

        if depends_on:
            self._interdeps = self._interdeps.extend(
                                  dependencies={paramspec: depends_on})
        if inf_from:
            self._interdeps = self._interdeps.extend(
                                  inferences={paramspec: inf_from})
        if not(depends_on or inf_from):
            self._interdeps = self._interdeps.extend(standalones=(paramspec,))

        log.info(f'Registered {name} in the Measurement.')

        return self

    def _register_arrayparameter(self,
                                 parameter: ArrayParameter,
                                 setpoints: Optional[setpoints_type],
                                 basis: Optional[setpoints_type],
                                 paramtype: str, ) -> None:
        """
        Register an ArrayParameter and the setpoints belonging to that
        ArrayParameter
        """
        my_setpoints = list(setpoints) if setpoints else []
        for i in range(len(parameter.shape)):
            if parameter.setpoint_full_names is not None and \
                    parameter.setpoint_full_names[i] is not None:
                spname = parameter.setpoint_full_names[i]
            else:
                spname = f'{parameter.full_name}_setpoint_{i}'
            if parameter.setpoint_labels:
                splabel = parameter.setpoint_labels[i]
            else:
                splabel = ''
            if parameter.setpoint_units:
                spunit = parameter.setpoint_units[i]
            else:
                spunit = ''

            self._register_parameter(name=spname,
                                     paramtype=paramtype,
                                     label=splabel,
                                     unit=spunit,
                                     setpoints=None,
                                     basis=None)

            my_setpoints += [spname]

        self._register_parameter(parameter.full_name,
                                 parameter.label,
                                 parameter.unit,
                                 my_setpoints,
                                 basis,
                                 paramtype)

    def _register_parameter_with_setpoints(self,
                                           parameter: ParameterWithSetpoints,
                                           setpoints: Optional[setpoints_type],
                                           basis: Optional[setpoints_type],
                                           paramtype: str) -> None:
        """
        Register an ParameterWithSetpoints and the setpoints belonging to the
        Parameter
        """
        my_setpoints = list(setpoints) if setpoints else []
        for sp in parameter.setpoints:
            if not isinstance(sp, Parameter):
                raise RuntimeError("The setpoints of a "
                                   "ParameterWithSetpoints "
                                   "must be a Parameter")
            spname = sp.full_name
            splabel = sp.label
            spunit = sp.unit

            self._register_parameter(name=spname,
                                     paramtype=paramtype,
                                     label=splabel,
                                     unit=spunit,
                                     setpoints=None,
                                     basis=None)

            my_setpoints.append(spname)

        self._register_parameter(parameter.full_name,
                                 parameter.label,
                                 parameter.unit,
                                 my_setpoints,
                                 basis,
                                 paramtype)

    def _register_multiparameter(self,
                                 multiparameter: MultiParameter,
                                 setpoints: Optional[setpoints_type],
                                 basis: Optional[setpoints_type],
                                 paramtype: str) -> None:
        """
        Find the individual multiparameter components and their setpoints
        and register those as individual parameters
        """
        setpoints_lists = []
        for i in range(len(multiparameter.shapes)):
            shape = multiparameter.shapes[i]
            name = multiparameter.full_names[i]
            if shape == ():
                my_setpoints = setpoints
            else:
                my_setpoints = list(setpoints) if setpoints else []
                for j in range(len(shape)):
                    if multiparameter.setpoint_full_names is not None and \
                            multiparameter.setpoint_full_names[i] is not None:
                        spname = multiparameter.setpoint_full_names[i][j]
                    else:
                        spname = f'{name}_setpoint_{j}'
                    if multiparameter.setpoint_labels is not None and \
                            multiparameter.setpoint_labels[i] is not None:
                        splabel = multiparameter.setpoint_labels[i][j]
                    else:
                        splabel = ''
                    if multiparameter.setpoint_units is not None and \
                            multiparameter.setpoint_units[i] is not None:
                        spunit = multiparameter.setpoint_units[i][j]
                    else:
                        spunit = ''

                    self._register_parameter(name=spname,
                                             paramtype=paramtype,
                                             label=splabel,
                                             unit=spunit,
                                             setpoints=None,
                                             basis=None)

                    my_setpoints += [spname]

            setpoints_lists.append(my_setpoints)

        for i, setpoints in enumerate(setpoints_lists):
            self._register_parameter(multiparameter.full_names[i],
                                     multiparameter.labels[i],
                                     multiparameter.units[i],
                                     setpoints,
                                     basis,
                                     paramtype)

    def register_custom_parameter(
            self: T, name: str,
            label: Optional[str] = None, unit: Optional[str] = None,
            basis: Optional[setpoints_type] = None,
            setpoints: Optional[setpoints_type] = None,
            paramtype: str = 'numeric') -> T:
        """
        Register a custom parameter with this measurement

        Args:
            name: The name that this parameter will have in the dataset. Must
                be unique (will overwrite an existing parameter with the same
                name!)
            label: The label
            unit: The unit
            basis: A list of either QCoDeS Parameters or the names
                of parameters already registered in the measurement that
                this parameter is inferred from
            setpoints: A list of either QCoDeS Parameters or the names of
                of parameters already registered in the measurement that
                are the setpoints of this parameter
            paramtype: Type of the parameter, i.e. the SQL storage class
        """
        return self._register_parameter(name,
                                        label,
                                        unit,
                                        setpoints,
                                        basis,
                                        paramtype)

    def unregister_parameter(self,
                             parameter: setpoints_type) -> None:
        """
        Remove a custom/QCoDeS parameter from the dataset produced by
        running this measurement
        """
        if isinstance(parameter, _BaseParameter):
            param = str(parameter)
        elif isinstance(parameter, str):
            param = parameter
        else:
            raise ValueError('Wrong input type. Must be a QCoDeS parameter or'
                             ' the name (a string) of a parameter.')

        try:
            paramspec: ParamSpecBase = self._interdeps[param]
        except KeyError:
            return

        self._interdeps = self._interdeps.remove(paramspec)

        log.info(f'Removed {param} from Measurement.')

    def add_before_run(self: T, func: Callable[..., Any], args: Sequence[Any]) -> T:
        """
        Add an action to be performed before the measurement.

        Args:
            func: Function to be performed
            args: The arguments to said function
        """
        # some tentative cheap checking
        nargs = len(signature(func).parameters)
        if len(args) != nargs:
            raise ValueError('Mismatch between function call signature and '
                             'the provided arguments.')

        self.enteractions.append((func, args))

        return self

    def add_after_run(self: T,
                      func: Callable[..., Any], args: Sequence[Any]) -> T:
        """
        Add an action to be performed after the measurement.

        Args:
            func: Function to be performed
            args: The arguments to said function
        """
        # some tentative cheap checking
        nargs = len(signature(func).parameters)
        if len(args) != nargs:
            raise ValueError('Mismatch between function call signature and '
                             'the provided arguments.')

        self.exitactions.append((func, args))

        return self

    def add_subscriber(
            self: T,
            func: Callable[..., Any],
            state: Union[MutableSequence[Any], MutableMapping[Any, Any]]
    ) -> T:
        """
        Add a subscriber to the dataset of the measurement.

        Args:
            func: A function taking three positional arguments: a list of
                tuples of parameter values, an integer, a mutable variable
                (list or dict) to hold state/writes updates to.
            state: The variable to hold the state.
        """
        self.subscribers.append((func, state))

        return self

    def set_shapes(self, shapes: Shapes) -> None:
        """
        Set the shapes of the data to be recorded in this
        measurement.

        Args:
            shapes: Dictionary from names of dependent parameters to a tuple
                of integers describing the shape of the measurement.
        """
        RunDescriber._verify_interdeps_shape(interdeps=self._interdeps,
                                             shapes=shapes)
        self._shapes = shapes

    def run(
        self,
        write_in_background: Optional[bool] = None,
        in_memory_cache: bool = True,
        dataset_class: Type[DataSetProtocol] = DataSet,
    ) -> Runner:
        """
        Returns the context manager for the experimental run

        Args:
            write_in_background: if True, results that will be added
                within the context manager with ``DataSaver.add_result``
                will be stored in background, without blocking the
                main thread that is executing the context manager.
                By default the setting for write in background will be
                read from the ``qcodesrc.json`` config file.
            in_memory_cache: Should measured data be keep in memory
                and available as part of the `dataset.cache` object.
            dataset_class: Class implementing the dataset protocol interface
                used to store the data.
        """
        if write_in_background is None:
            write_in_background = qc.config.dataset.write_in_background
        return Runner(
            self.enteractions,
            self.exitactions,
            self.experiment,
            station=self.station,
            write_period=self._write_period,
            interdeps=self._interdeps,
            name=self.name,
            subscribers=self.subscribers,
            parent_datasets=self._parent_datasets,
            extra_log_info=self._extra_log_info,
            write_in_background=write_in_background,
            shapes=self._shapes,
            in_memory_cache=in_memory_cache,
            dataset_class=dataset_class,
        )



from typing import Any, Dict, Optional
import numpy as np

from qcodes.instrument.visa import VisaInstrument
from qcodes.instrument.parameter import Parameter, DelegateParameter
from qcodes.utils import validators as vals


class SP983A(VisaInstrument):
    """
    A driver for Basel Preamp's (SP983C) Remote Instrument - Model SP983A.

    Args:
        name: name for your instrument driver instance
        address: address of the connected remote controller of basel preamp
        input_offset_voltage: (Optional) A source input offset voltage
            parameter. The range for input is -10 to 10 Volts and it is
            user's responsibility to ensure this. This source parameter is
            used to set offset voltage parameter of the preamp and the
            source parameter should represent a voltage source that is
            connected to the "Offset Input Volgate" connector of the SP983C.
    """
    def __init__(self,
                 name: str,
                 address: str,
                 input_offset_voltage: Optional[Parameter] = None,
                 terminator: str = "\r\n",
                 **kwargs: Any) -> None:
        super().__init__(name, address, terminator=terminator, **kwargs)

        self.connect_message()

        self.add_parameter(
            "gain",
            label="Gain",
            unit="V/A",
            set_cmd=self._set_gain,
            get_cmd=self._get_gain,
            vals=vals.Enum(1e5, 1e6, 1e7, 1e8, 1e9),
        )
        self.add_parameter(
            "fcut",
            unit="Hz",
            label="Filter Cut-Off Frequency",
            get_cmd=self._get_filter,
            get_parser=self._parse_filter_value,
            set_cmd=self._set_filter,
            val_mapping={30: '30', 100: '100', 300: '300', 1000: '1k',
                         3000: '3k', 10e3: '10k', 30e3: '30k',
                         100e3: '100k', 1e6: 'FULL'},
        )
        self.add_parameter(
            "overload_status",
            label="Overload Status",
            set_cmd=False,
            get_cmd="GET O"
        )

        self.add_parameter(
            "offset_voltage",
            label="Offset Voltage for SP983C",
            unit="V",
            vals=vals.Numbers(-0.1, 0.1),
            scale=100,
            source=input_offset_voltage,
            parameter_class=DelegateParameter)

    def get_idn(self) -> Dict[str, Optional[str]]:
        vendor = 'Physics Basel'
        model = 'SP 983A'
        serial = None
        firmware = None
        return {'vendor': vendor, 'model': model,
                'serial': serial, 'firmware': firmware}

    def _set_gain(self, value: float) -> None:
        r = self.ask(f"SET G 1E{int(np.log10(value))}")
        if r != "OK":
            raise ValueError(f"Expected OK return but got: {r}")

    def _get_gain(self) -> float:
        s = self.ask("GET G")
        r = s.split("Gain: ")[1]
        return float(r)

    def _set_filter(self, value: str) -> None:
        r = self.ask(f"SET F {value}")
        if r != "OK":
            raise ValueError(f"Expected OK return but got: {r}")

    def _get_filter(self) -> str:
        s = self.ask("GET F")
        return s.split("Filter: ")[1]

    @staticmethod
    def _parse_filter_value(val: str) -> str:
        if val.startswith("F"):
            return "FULL"
        elif val[-3::] == "kHz":
            return str(int(val[0:-3]))+'k'

        elif val[-2::] == "Hz":
            return str(int(val[0:-2]))
        else:
            raise ValueError(f"Could not interpret result. Got: {val}")



"""
This module implements a :class:`.Group` intended to hold multiple
parameters that are to be gotten and set by the same command. The parameters
should be of type :class:`GroupParameter`
"""


from collections import OrderedDict
from typing import Any, Callable, Dict, Mapping, Optional, Sequence, Union

from qcodes.instrument.base import InstrumentBase
from qcodes.instrument.parameter import ParamDataType, Parameter, ParamRawDataType


class GroupParameter(Parameter):
    """
    Group parameter is a :class:`.Parameter`, whose value can be set or get
    only with other group parameters. This happens when an instrument
    has commands which set and get more than one parameter per call.

    The ``set_raw`` method of a group parameter forwards the call to the
    group, and the group then makes sure that the values of other parameters
    within the group are left unchanged. The ``get_raw`` method of a group
    parameter also forwards the call to the group, and the group makes sure
    that the command output is parsed correctly, and the value of the
    parameter of interest is returned.

    After initialization, the group parameters need to be added to a group.
    See :class:`.Group` for more information.

    Args:
        name: Name of the parameter.
        instrument: Instrument that this parameter belongs to; this instrument
            is used by the group to call its get and set commands.
        initial_value: Initial value of the parameter. Note that either none or
            all of the parameters in a :class:`.Group` should have an initial
            value.

        **kwargs: All kwargs used by the :class:`.Parameter` class, except
             ``set_cmd`` and ``get_cmd``.
    """

    def __init__(self,
                 name: str,
                 instrument: Optional['InstrumentBase'] = None,
                 initial_value: Union[float, int, str, None] = None,
                 **kwargs: Any
                 ) -> None:

        if "set_cmd" in kwargs or "get_cmd" in kwargs:
            raise ValueError("A GroupParameter does not use 'set_cmd' or "
                             "'get_cmd' kwarg")

        self._group: Union[Group, None] = None
        self._initial_value = initial_value
        super().__init__(name, instrument=instrument, **kwargs)

    @property
    def group(self) -> Optional['Group']:
        """
        The group that this parameter belongs to.
        """
        return self._group

    def get_raw(self) -> ParamRawDataType:
        if self.group is None:
            raise RuntimeError("Trying to get Group value but no "
                               "group defined")
        self.group.update()
        return self.cache.raw_value

    def set_raw(self, value: ParamRawDataType) -> None:
        if self.group is None:
            raise RuntimeError("Trying to set Group value but no "
                               "group defined")
        self.group._set_one_parameter_from_raw(self, value)


class Group:
    """
    The group combines :class:`.GroupParameter` s that are to be gotten or set
    via the same command. The command has to be a string, for example,
    a VISA command.

    The :class:`Group`'s methods are used within :class:`GroupParameter` in
    order to properly implement setting and getting of a single parameter in
    the situation where one command sets or gets more than one parameter.

    The command used for setting values of parameters has to be a format
    string which contains the names of the parameters the group has been
    initialized with. For example, if a command has syntax ``CMD a_value,
    b_value``, where ``a_value`` and ``b_value`` are values of two parameters
    with names ``a`` and ``b``, then the command string has to be ``CMD {a},
    {b}``, and the group has to be initialized with two ``GroupParameter`` s
    ``a_param`` and ``b_param``, where ``a_param.name=="a"`` and
    ``b_param.name=="b"``.

    **Note** that by default, it is assumed that the command used for getting
    values returns a comma-separated list of values of parameters, and their
    order corresponds to the order of :class:`.GroupParameter` s in the list
    that is passed to the :class:`Group`'s constructor. Through keyword
    arguments of the :class:`Group`'s constructor, it is possible to change
    the separator, and even the parser of the output of the get command.

    The get and set commands are called via the instrument that the first
    parameter belongs to. It is assumed that all the parameters within the
    group belong to the same instrument.

    Example:

        ::

            class InstrumentWithGroupParameters(VisaInstrument):
                def __init__(self, name, address, **kwargs):
                    super().__init__(name, address, **kwargs)

                    ...

                    # Here is how group of group parameters is defined for
                    # a simple case of an example "SGP" command that sets and gets
                    # values of "enabled" and "gain" parameters (it is assumed that
                    # "SGP?" returns the parameter values as comma-separated list
                    # "enabled_value,gain_value")
                    self.add_parameter('enabled',
                                       label='Enabled',
                                       val_mapping={True: 1, False: 0},
                                       parameter_class=GroupParameter)
                    self.add_parameter('gain',
                                       label='Some gain value',
                                       get_parser=float,
                                       parameter_class=GroupParameter)
                    self.output_group = Group([self.enabled, self.gain],
                                              set_cmd='SGP {enabled}, {gain}',
                                              get_cmd='SGP?')

                    ...

    Args:
        parameters: a list of :class:`.GroupParameter` instances which have
            to be gotten and set via the same command; the order of
            parameters in the list should correspond to the order of the
            values returned by the ``get_cmd``.
        set_cmd: Format string of the command that is used for setting the
            values of the parameters; for example, ``CMD {a}, {b}``.
        get_cmd: String of the command that is used for getting the values
            of the parameters; for example, ``CMD?``.
        separator: A separator that is used when parsing the output of the
            ``get_cmd`` in order to obtain the values of the parameters; it
            is ignored in case a custom ``get_parser`` is used.
        get_parser: A callable with a single string argument that is used to
            parse the output of the ``get_cmd``; the callable has to return a
            dictionary where parameter names are keys, and the values are the
            values (as directly obtained from the output of the get command;
            note that parsers within the parameters will take care of
            individual parsing of their values).
        single_instrument: A flag to indicate that all parameters belong to a
        single instrument, which in turn does additional checks. Defaults to True.
    """

    def __init__(
        self,
        parameters: Sequence[GroupParameter],
        set_cmd: Optional[str] = None,
        get_cmd: Optional[str] = None,
        get_parser: Union[Callable[[str], Mapping[str, Any]], None] = None,
        separator: str = ",",
        single_instrument: bool = True,
    ) -> None:
        self._parameters = OrderedDict((p.name, p) for p in parameters)

        for p in parameters:
            p._group = self

        if single_instrument:
            if len({p.root_instrument for p in parameters}) > 1:
                raise ValueError(
                    "All parameters should belong to the same instrument")

        self._instrument = parameters[0].root_instrument

        self._set_cmd = set_cmd
        self._get_cmd = get_cmd

        if get_parser:
            self.get_parser = get_parser
        else:
            self.get_parser = self._separator_parser(separator)

        if single_instrument:
            self._check_initial_values(parameters)

    def _check_initial_values(self, parameters: Sequence[GroupParameter]) -> None:
        have_initial_values = [p._initial_value is not None
                               for p in parameters]
        if any(have_initial_values):
            if not all(have_initial_values):
                params_with_initial_values = [p.name for p in parameters
                                              if p._initial_value is not None]
                params_without_initial_values = [p.name for p in parameters
                                                 if p._initial_value is None]
                error_msg = (f'Either none or all of the parameters in a '
                             f'group should have an initial value. Found '
                             f'initial values for '
                             f'{params_with_initial_values} but not for '
                             f'{params_without_initial_values}.')
                raise ValueError(error_msg)

            calling_dict = {name: p._from_value_to_raw_value(p._initial_value)
                            for name, p in self.parameters.items()}

            self._set_from_dict(calling_dict)

    def _separator_parser(self, separator: str
                          ) -> Callable[[str], Dict[str, ParamRawDataType]]:
        """A default separator-based string parser"""
        def parser(ret_str: str) -> Dict[str, Any]:
            keys = self.parameters.keys()
            values = ret_str.split(separator)
            return dict(zip(keys, values))

        return parser

    def set_parameters(self, parameters_dict: Mapping[str, ParamDataType]) -> None:
        """
        Sets the value of one or more parameters within a group to the given
        values by calling the ``set_cmd`` while updating rest.

        Args:
            parameters_dict: The dictionary of one or more parameters within
            the group with the corresponding values to be set.
        """
        if not parameters_dict:
            raise RuntimeError("Provide at least one group parameter and its "
                               "value to be set.")
        if any((p.get_latest() is None) for p in self.parameters.values()):
            self.update()
        calling_dict = {name: p.cache.raw_value
                        for name, p in self.parameters.items()}
        for parameter_name, value in parameters_dict.items():
            p = self.parameters[parameter_name]
            raw_value = p._from_value_to_raw_value(value)
            calling_dict[parameter_name] = raw_value

        self._set_from_dict(calling_dict)

    def _set_one_parameter_from_raw(self, set_parameter: GroupParameter,
                                    raw_value: ParamRawDataType) -> None:
        """
        Sets the raw_value of the given parameter within a group to the given
        raw_value by calling the ``set_cmd``.

        Args:
            set_parameter: The parameter within the group to set.
            raw_value: The new raw_value for this parameter.
        """
        # TODO replace get latest with call to cache.invalid once that lands
        if any((p.get_latest() is None) for p in self.parameters.values()):
            self.update()
        calling_dict = {name: p.cache.raw_value
                        for name, p in self.parameters.items()}
        calling_dict[set_parameter.name] = raw_value

        self._set_from_dict(calling_dict)

    def _set_from_dict(self, calling_dict: Mapping[str, ParamRawDataType]) -> None:
        """
        Use ``set_cmd`` to parse a dict that maps parameter names to parameter
        raw values, and actually perform setting the values.
        """
        if self._set_cmd is None:
            raise RuntimeError("Calling set but no `set_cmd` defined")
        command_str = self._set_cmd.format(**calling_dict)
        if self.instrument is None:
            raise RuntimeError("Trying to set GroupParameter not attached "
                               "to any instrument.")
        self.instrument.write(command_str)
        for name, p in list(self.parameters.items()):
            p.cache._set_from_raw_value(calling_dict[name])

    def update(self) -> None:
        """
        Update the values of all the parameters within the group by calling
        the ``get_cmd``.
        """
        if self.instrument is None:
            raise RuntimeError("Trying to update GroupParameter not attached "
                               "to any instrument.")
        if self._get_cmd is None:
            parameter_names = ', '.join(
                p.full_name for p in self.parameters.values())
            raise RuntimeError(f'Cannot update values in the group with '
                               f'parameters - {parameter_names} since it '
                               f'has no `get_cmd` defined.')
        ret = self.get_parser(self.instrument.ask(self._get_cmd))
        for name, p in list(self.parameters.items()):
            p.cache._set_from_raw_value(ret[name])

    @property
    def parameters(self) -> "OrderedDict[str, GroupParameter]":
        """
        All parameters in this group as a dict from parameter name to
        :class:`.Parameter`
        """
        return self._parameters

    @property
    def instrument(self) -> Optional[InstrumentBase]:
        """
        The ``root_instrument`` that this parameter belongs to.
        """
        return self._instrument



"""
Sometimes it happens that databases are put into inconsistent/corrupt states.
This module contains functions to remedy known issues.
"""
import json
import logging
from typing import Any, Dict, Sequence, cast

from tqdm import tqdm

import qcodes.dataset.descriptions.versioning.serialization as serial
import qcodes.dataset.descriptions.versioning.v0 as v0
from qcodes.dataset.descriptions.rundescriber import RunDescriber
from qcodes.dataset.descriptions.versioning.converters import old_to_new
from qcodes.dataset.descriptions.versioning.rundescribertypes import \
    RunDescriberV1Dict
from qcodes.dataset.sqlite.connection import (ConnectionPlus, atomic,
                                              atomic_transaction)
from qcodes.dataset.sqlite.db_upgrades import get_user_version
from qcodes.dataset.sqlite.queries import (_get_parameters,
                                           _update_run_description,
                                           get_run_description,
                                           update_run_description)
from qcodes.dataset.sqlite.query_helpers import one, select_one_where

log = logging.getLogger(__name__)


def fix_version_4a_run_description_bug(conn: ConnectionPlus) -> Dict[str, int]:
    """
    Fix function to fix a bug where the RunDescriber accidentally wrote itself
    to string using the (new) InterDependencies_ object instead of the (old)
    InterDependencies object. After the first call, this function should be
    idempotent.


    Args:
        conn: the connection to the database

    Returns:
        A dict with the fix results ('runs_inspected', 'runs_fixed')
    """

    user_version = get_user_version(conn)

    if not user_version == 4:
        raise RuntimeError('Database of wrong version. Will not apply fix. '
                           'Expected version 4, found version {user_version}')

    no_of_runs_query = "SELECT max(run_id) FROM runs"
    no_of_runs = one(atomic_transaction(conn, no_of_runs_query), 'max(run_id)')
    no_of_runs = no_of_runs or 0

    with atomic(conn) as conn:

        pbar = tqdm(range(1, no_of_runs+1))
        pbar.set_description("Fixing database")

        # collect some metrics
        runs_inspected = 0
        runs_fixed = 0

        old_style_keys = ['paramspecs']
        new_style_keys = ['parameters', 'dependencies', 'inferences',
                          'standalones']

        for run_id in pbar:

            desc_str = get_run_description(conn, run_id)
            desc_ser = json.loads(desc_str)
            idps_ser = desc_ser['interdependencies']

            if list(idps_ser.keys()) == old_style_keys:
                pass
            elif list(idps_ser.keys()) == new_style_keys:
                old_desc_ser = \
                    _convert_run_describer_v1_like_dict_to_v0_like_dict(
                        desc_ser)
                json_str = json.dumps(old_desc_ser)
                _update_run_description(conn, run_id, json_str)
                runs_fixed += 1
            else:
                raise RuntimeError(f'Invalid runs_description for run_id: '
                                   f'{run_id}')

            runs_inspected += 1

    return {'runs_inspected': runs_inspected, 'runs_fixed': runs_fixed}


def _convert_run_describer_v1_like_dict_to_v0_like_dict(
        new_desc_dict: RunDescriberV1Dict) -> Dict[str, Any]:
    """
    This function takes the given dict which is expected to be
    representation of `RunDescriber` with `InterDependencies_` (underscore!)
    object and without "version" field, and converts it to a dict that is a
    representation of the `RunDescriber` object with `InterDependencies`
    (no underscore!) object and without "version" field.
    """
    new_desc_dict = new_desc_dict.copy()
    # We intend to use conversion methods from `serialization` module,
    # but those work only with RunDescriber representations that have
    # "version" field. So first, the "version" field with correct value is
    # added.
    new_desc_dict['version'] = 1
    # Out of that dict we create RunDescriber object of the current version
    # (regardless of what the current version is).
    new_desc = serial.from_dict_to_current(new_desc_dict)
    # The RunDescriber of the current version gets converted to a dictionary
    # that represents a RunDescriber object of version 0 - this is the one
    # that has InterDependencies object in it (not the InterDependencies_ one).
    old_desc_dict = cast(Dict[str, Any], serial.to_dict_as_version(new_desc, 0))
    # Lastly, the "version" field is removed.
    old_desc_dict.pop('version')
    return old_desc_dict


def fix_wrong_run_descriptions(conn: ConnectionPlus,
                               run_ids: Sequence[int]) -> None:
    """
    NB: This is a FIX function. Do not use it unless your database has been
    diagnosed with the problem that this function fixes.

    Overwrite faulty run_descriptions by using information from the layouts and
    dependencies tables. If a correct description is found for a run, that
    run is left untouched.

    Args:
        conn: The connection to the database
        run_ids: The runs to (potentially) fix
    """

    user_version = get_user_version(conn)

    if not user_version == 3:
        raise RuntimeError('Database of wrong version. Will not apply fix. '
                           'Expected version 3, found version {user_version}')


    log.info('[*] Fixing run descriptions...')
    for run_id in run_ids:
        trusted_paramspecs = _get_parameters(conn, run_id)
        interdeps = v0.InterDependencies(*trusted_paramspecs)
        interdeps_ = old_to_new(interdeps)
        trusted_desc = RunDescriber(interdeps_)

        actual_desc_str = select_one_where(conn, "runs",
                                           "run_description",
                                           "run_id", run_id)

        trusted_json = serial.to_json_as_version(trusted_desc, 0)

        if actual_desc_str == trusted_json:
            log.info(f'[+] Run id: {run_id} had an OK description')
        else:
            log.info(f'[-] Run id: {run_id} had a broken description. '
                     f'Description found: {actual_desc_str}')
            update_run_description(conn, run_id, trusted_json)
            log.info(f'    Run id: {run_id} has been updated.')



from typing import Any, Tuple, List
import logging
from collections import namedtuple
from itertools import chain

log = logging.getLogger(__name__)

try:
    import win32com.client
    import pythoncom
    from pythoncom import VT_BYREF, VT_R8, VT_I4
except ImportError as e:
    message = ('To use the DynaCool Driver, please install win32com.'
               ' Installation can be done with pip install pypiwin32com')
    log.exception(message)
    raise ImportError(message)


CmdArgs = namedtuple('CmdArgs', 'cmd args')

# The length of a command header, aka a command keyword
# Every command sent from the driver via the server must have a
# keyword of exactly this length ('?' NOT included)
CMD_HEADER_LENGTH = 4


class CommandHandler:
    """
    This is the class that gets called by the server.py

    This class is responsible for making the actual calls into the instrument
    firmware. The idea is that this class get a SCPI-like string from the
    server, e.g. 'TEMP?' or 'TEMP 300, 10, 1' and then makes the corresponding
    MultiVu API call (or returns an error message to the server).
    """

    # Variable types
    _variants = {'double': win32com.client.VARIANT(VT_BYREF | VT_R8, 0.0),
                 'long': win32com.client.VARIANT(VT_BYREF | VT_I4, 0)}

    def __init__(self, inst_type: str = 'dynacool') -> None:
        self.inst_type = inst_type
        pythoncom.CoInitialize()
        client_id = f'QD.MULTIVU.{inst_type.upper()}.1'
        try:
            self._mvu = win32com.client.Dispatch(client_id)
        except pythoncom.com_error:
            error_mssg = ('Could not connect to Multivu Application. Please '
                          'make sure that the MultiVu Application is running.')
            log.exception(error_mssg)
            raise ValueError(error_mssg)

        _variants = CommandHandler._variants

        # Hard-code what we know about the MultiVu API
        self._gets = {'TEMP': CmdArgs(cmd=self._mvu.GetTemperature,
                                      args=[_variants['double'],
                                            _variants['long']]),
                      'CHAT': CmdArgs(cmd=self._mvu.GetChamberTemp,
                                      args=[_variants['double'],
                                            _variants['long']]),
                      'GLTS': CmdArgs(cmd=self._mvu.GetLastTempSetpoint,
                                      args=[_variants['double'],
                                            _variants['double'],
                                            _variants['long']]),
                      'GLFS': CmdArgs(cmd=self._mvu.GetFieldSetpoints,
                                      args=[_variants['double'],
                                            _variants['double'],
                                            _variants['long'],
                                            _variants['long']]),
                      'CHAM': CmdArgs(cmd=self._mvu.GetChamber,
                                      args=[_variants['long']]),
                      'FELD': CmdArgs(cmd=self._mvu.GetField,
                                      args=[_variants['double'],
                                            _variants['long']]),
                      '*IDN': CmdArgs(cmd=self.make_idn_string, args=[])}

        self._sets = {'TEMP': self._mvu.SetTemperature,
                      'FELD': self._mvu.SetField}

        # validate the commands
        for cmd in chain(self._gets, self._sets):
            if len(cmd) != CMD_HEADER_LENGTH:
                raise ValueError(f'Invalid command length: {cmd}.'
                                 f' Must have length {CMD_HEADER_LENGTH}')

    def make_idn_string(self) -> str:
        return f'0, QuantumDesign, {self.inst_type}, N/A, N/A'

    def preparser(self, cmd_str: str) -> Tuple[CmdArgs, bool]:
        """
        Parse the raw SCPI-like input string into a CmdArgs tuple containing
        the corresponding MultiVu API function and a boolean indicating whether
        we expect the MultiVu function to modify its input (i.e. be a query)

        Args:
            cmd_str: A SCPI-like string, e.g. 'TEMP?' or 'TEMP 300, 0.1, 1'

        Returns:
            A tuple of a CmdArgs tuple and a bool indicating whether this was
            a query
        """
        def err_func() -> int:
            return -2

        cmd_head = cmd_str[:CMD_HEADER_LENGTH]

        if cmd_head not in set(self._gets.keys()).union(set(self._sets.keys())):
            cmd = err_func
            args: List[Any] = []
            is_query = False

        elif cmd_str.endswith('?'):
            cmd = self._gets[cmd_head].cmd
            args = self._gets[cmd_head].args
            is_query = True
        else:
            cmd = self._sets[cmd_head]
            args = list(float(arg) for arg in cmd_str[5:].split(', '))
            is_query = False

        return CmdArgs(cmd=cmd, args=args), is_query

    @staticmethod
    def postparser(error_code: int, vals: List[Any]) -> str:
        """
        Parse the output of the MultiVu API call into a string that the server
        can send back to the client

        Args:
            error_code: the error code returned from the MultiVu call
            vals: A list of the returned values (empty in case of a set cmd)
        """
        response = f'{error_code}'

        for val in vals:
            response += f', {val}'

        return response

    def __call__(self, cmd: str) -> str:

        cmd_and_args, is_query = self.preparser(cmd)
        log.debug(f'Parsed {cmd} into {cmd_and_args}')

        # Actually perform the call into the MultiVu API
        error_code = cmd_and_args.cmd(*cmd_and_args.args)

        # return values in case we did a query
        if is_query:
            # read out the mutated values
            # (win32 reverses the order)
            vals = list(arg.value for arg in cmd_and_args.args)
            vals.reverse()
            # reset the value variables for good measures
            for arg in cmd_and_args.args:
                arg.value = 0
        else:
            vals = []

        response_message = self.postparser(error_code, vals)

        return response_message



import collections
from typing import Any, Dict, Optional

import numpy as np
import xarray as xr

from qcodes.utils.helpers import DelegateAttributes, full_class, warn_units


class DataArray(DelegateAttributes):

    """
    A container for one parameter in a measurement loop.

    If this is a measured parameter, This object doesn't contain
    the data of the setpoints it was measured at, but it references
    the DataArray objects of these parameters. Those objects only have
    the dimensionality at which they were set - ie the inner loop setpoint
    the same dimensionality as the measured parameter, but the outer
    loop setpoint(s) have lower dimensionality

    When it's first created, a DataArray has no dimensionality, you must call
    .nest for each dimension.

    If preset_data is provided it is used to initialize the data, and the array
    can still be nested around it (making many copies of the data).
    Otherwise it is an error to nest an array that already has data.

    Once the array is initialized, a DataArray acts a lot like a numpy array,
    because we delegate attributes through to the numpy array

    Args:
        parameter (Optional[Parameter]): The parameter whose values will
            populate this array, if any. Will copy ``name``, ``full_name``,
            ``label``, ``unit``, and ``snapshot`` from here unless you
            provide them explicitly.

        name (Optional[str]): The short name of this array.
            TODO: use full_name as name, and get rid of short name

        full_name (Optional[str]): The complete name of this array. If the
            array is based on a parameter linked to an instrument, this is
            typically '<instrument_name>_<param_name>'

        label (Optional[str]): A description of the values in this array to
            use for axis and colorbar labels on plots.

        snapshot (Optional[dict]): Metadata snapshot to save with this array.

        array_id (Optional[str]): A name for this array that's unique within
            its ``DataSet``. Typically the full_name, but when the ``DataSet``
            is constructed we will append '_<i>' (``i`` is an integer starting
            from 1) if necessary to differentiate arrays with the same id.
            TODO: this only happens for arrays provided to the DataSet
            constructor, not those added with add_array. Fix this!
            Also, do we really need array_id *and* full_name (let alone name
            but I've already said we should remove this)?

        set_arrays (Optional[Tuple[DataArray]]): If this array is being
            created with shape already, you can provide one setpoint array
            per dimension. The first should have one dimension, the second
            two dimensions, etc.

        shape (Optional[Tuple[int]]): The shape (as in numpy) of the array.
            Will be prepended with new dimensions by any calls to ``nest``.

        action_indices (Optional[Tuple[int]]): If used within a ``Loop``,
            these are the indices at each level of nesting within the
            ``Loop`` of the loop action that's populating this array.
            TODO: this shouldn't be in DataArray at all, the loop should
            handle converting this to array_id internally (maybe it
            already does?)

        unit (Optional[str]): The unit of the values stored in this array.

        units (Optional[str]): DEPRECATED, redirects to ``unit``.

        is_setpoint (bool): True if this is a setpoint array, False if it
            is measured. Default False.

        preset_data (Optional[Union[numpy.ndarray, Sequence]]): Contents of the
            array, if already known (for example if this is a setpoint
            array). ``shape`` will be inferred from this array instead of
            from the ``shape`` argument.
    """

    # attributes of self to include in the snapshot
    SNAP_ATTRS = (
        'array_id',
        'name',
        'shape',
        'unit',
        'label',
        'action_indices',
        'is_setpoint')

    # attributes of the parameter (or keys in the incoming snapshot)
    # to copy to DataArray attributes, if they aren't set some other way
    COPY_ATTRS_FROM_INPUT = (
        'name',
        'label',
        'unit')

    # keys in the parameter snapshot to omit from our snapshot
    SNAP_OMIT_KEYS = (
        'ts',
        'value',
        '__class__',
        'set_arrays',
        'shape',
        'array_id',
        'action_indices')

    def __init__(self, parameter=None, name=None, full_name=None, label=None,
                 snapshot=None, array_id=None, set_arrays=(), shape=None,
                 action_indices=(), unit=None, units=None, is_setpoint=False,
                 preset_data=None):
        self.name = name
        self.full_name = full_name or name
        self.label = label
        self.shape = shape
        if units is not None:
            warn_units('DataArray', self)
            if unit is None:
                unit = units
        self.unit = unit
        self.array_id = array_id
        self.is_setpoint = is_setpoint
        self.action_indices = action_indices
        self.set_arrays = set_arrays

        self._preset = False

        # store a reference up to the containing DataSet
        # this also lets us make sure a DataArray is only in one DataSet
        self._data_set = None

        self.last_saved_index = None
        self.modified_range = None

        self.ndarray = None
        if snapshot is None:
            snapshot = {}
        self._snapshot_input = {}

        if parameter is not None:
            param_full_name = getattr(parameter, 'full_name', None)
            if param_full_name and not full_name:
                self.full_name = parameter.full_name

            if hasattr(parameter, 'snapshot') and not snapshot:
                snapshot = parameter.snapshot()
            else:
                # TODO: why is this in an else clause?
                for attr in self.COPY_ATTRS_FROM_INPUT:
                    if (hasattr(parameter, attr) and
                            not getattr(self, attr, None)):
                        setattr(self, attr, getattr(parameter, attr))

        for key, value in snapshot.items():
            if key not in self.SNAP_OMIT_KEYS:
                self._snapshot_input[key] = value

                if (key in self.COPY_ATTRS_FROM_INPUT and
                        not getattr(self, key, None)):
                    setattr(self, key, value)

        if not self.label:
            self.label = self.name

        if preset_data is not None:
            self.init_data(preset_data)
        elif shape is None:
            self.shape = ()

    @property
    def data_set(self):
        """
        The DataSet this array belongs to.

        A DataArray can belong to at most one DataSet.
        TODO: make this a weakref
        """
        return self._data_set

    @data_set.setter
    def data_set(self, new_data_set):
        if (self._data_set is not None and
                new_data_set is not None and
                self._data_set != new_data_set):
            raise RuntimeError('A DataArray can only be part of one DataSet')
        self._data_set = new_data_set

    def nest(self, size, action_index=None, set_array=None):
        """
        Nest this array inside a new outer loop.

        You cannot call ``nest`` after ``init_data`` unless this is a
        setpoint array.
        TODO: is this restriction really useful? And should we maintain
        a distinction between _preset and is_setpoint, or can wejust use
        is_setpoint?

        Args:
            size (int): Length of the new loop.

            action_index (Optional[int]): Within the outer loop at this
                nesting level, which action does this array derive from?

            set_array (Optional[DataArray]): The setpoints of the new outer
                loop. If this DataArray *is* a setpoint array, you should
                omit both ``action_index`` and ``set_array``, and it will
                reference itself as the inner setpoint array.

        Returns:
            DataArray: self, in case you want to construct the array with
                chained method calls.
        """
        if self.ndarray is not None and not self._preset:
            raise RuntimeError('Only preset arrays can be nested after data '
                               'is initialized! {}'.format(self))

        if set_array is None:
            if self.set_arrays:
                raise TypeError('a setpoint array must be its own inner loop')
            set_array = self

        self.shape = (size, ) + self.shape

        if action_index is not None:
            self.action_indices = (action_index, ) + self.action_indices

        self.set_arrays = (set_array, ) + self.set_arrays

        if self._preset:
            inner_data = self.ndarray
            self.ndarray = np.ndarray(self.shape)
            # existing preset array copied to every index of the nested array.
            for i in range(size):
                self.ndarray[i] = inner_data

            # update modified_range so the entire array still looks modified
            self.modified_range = (0, self.ndarray.size - 1)

            self._set_index_bounds()

        return self

    def init_data(self, data=None):
        """
        Create the actual numpy array to hold data.

        The array will be sized based on either ``self.shape`` or
        data provided here.

        Idempotent: will do nothing if the array already exists.

        If data is provided, this array is marked as a preset
        meaning it can still be nested around this data.
        TODO: per above, perhaps remove this distinction entirely?

        Args:
            data (Optional[Union[numpy.ndarray, Sequence]]): If provided,
                we fill the array with this data. Otherwise the new
                array will be filled with NaN.

        Raises:
            ValueError: if ``self.shape`` does not match ``data.shape``
            ValueError: if the array was already initialized with a
                different shape than we're about to create
        """
        if data is not None:
            if not isinstance(data, np.ndarray):
                if isinstance(data, collections.abc.Iterator):
                    # faster than np.array(tuple(data)) (or via list)
                    # but requires us to assume float
                    data = np.fromiter(data, float)
                else:
                    data = np.array(data)

            if self.shape is None:
                self.shape = data.shape
            elif data.shape != self.shape:
                raise ValueError('preset data must be a sequence '
                                 'with shape matching the array shape',
                                 data.shape, self.shape)
            self.ndarray = data
            self._preset = True

            # mark the entire array as modified
            self.modified_range = (0, data.size - 1)

        elif self.ndarray is not None:
            if self.ndarray.shape != self.shape:
                raise ValueError('data has already been initialized, '
                                 'but its shape doesn\'t match self.shape')
            return
        else:
            self.ndarray = np.ndarray(self.shape)
            self.clear()
        self._set_index_bounds()

    def _set_index_bounds(self):
        self._min_indices = [0 for d in self.shape]
        self._max_indices = [d - 1 for d in self.shape]

    def clear(self):
        """Fill the (already existing) data array with nan."""
        # only floats can hold nan values. I guess we could
        # also raise an error in this case? But generally float is
        # what people want anyway.
        if self.ndarray.dtype != float:
            self.ndarray = self.ndarray.astype(float)
        self.ndarray.fill(float('nan'))

    def __setitem__(self, loop_indices, value):
        """
        Set data values.

        Follows numpy syntax, allowing indices of lower dimensionality than
        the array, if value makes up the extra dimension(s)

        Also update the record of modifications to the array. If you don't
        want this overhead, you can access ``self.ndarray`` directly.
        """
        if isinstance(loop_indices, collections.abc.Iterable):
            min_indices = list(loop_indices)
            max_indices = list(loop_indices)
        else:
            min_indices = [loop_indices]
            max_indices = [loop_indices]

        for i, index in enumerate(min_indices):
            if isinstance(index, slice):
                start, stop, step = index.indices(self.shape[i])
                min_indices[i] = start
                max_indices[i] = start + (
                    ((stop - start - 1)//step) * step)

        min_li = self.flat_index(min_indices, self._min_indices)
        max_li = self.flat_index(max_indices, self._max_indices)
        self._update_modified_range(min_li, max_li)

        self.ndarray.__setitem__(loop_indices, value)

    def __getitem__(self, loop_indices):
        return self.ndarray[loop_indices]

    delegate_attr_objects = ['ndarray']

    def __len__(self):
        """
        Array length.

        Must be explicitly delegated, because len() will look for this
        attribute to already exist.
        """
        return len(self.ndarray)

    def flat_index(self, indices, index_fill=None):
        """
        Generate the raveled index for the given indices.

        This is the index you would have if the array is reshaped to 1D,
        looping over the indices from inner to outer.

        Args:
            indices (Sequence): indices of an element or slice of this array.

            index_fill (Optional[Sequence]): extra indices to use if
                ``indices`` has less dimensions than the array, ie it points
                to a slice rather than a single element. Use zeros to get the
                beginning of this slice, and [d - 1 for d in shape] to get the
                end of the slice.

        Returns:
            int: the resulting flat index.
        """
        if len(indices) < len(self.shape):
            indices = indices + index_fill[len(indices):]
        return np.ravel_multi_index(tuple(zip(indices)), self.shape)[0]

    def _update_modified_range(self, low, high):
        if self.modified_range:
            self.modified_range = (min(self.modified_range[0], low),
                                   max(self.modified_range[1], high))
        else:
            self.modified_range = (low, high)

    def mark_saved(self, last_saved_index):
        """
        Mark certain outstanding modifications as saved.

        Args:
            last_saved_index (int): The flat index of the last point
                saved. If ``modified_range`` extends beyond this, the
                data past ``last_saved_index`` will still be marked
                modified, otherwise ``modified_range`` is cleared
                entirely.
        """
        if self.modified_range:
            if last_saved_index >= self.modified_range[1]:
                self.modified_range = None
            else:
                self.modified_range = (max(self.modified_range[0],
                                           last_saved_index + 1),
                                       self.modified_range[1])
        self.last_saved_index = last_saved_index

    def clear_save(self):
        """
        Make previously saved parts of this array look unsaved (modified).

        This can be used to force overwrite or rewrite, like if we're
        moving or copying the ``DataSet``.
        """
        if self.last_saved_index is not None:
            self._update_modified_range(0, self.last_saved_index)

        self.last_saved_index = None

    def get_synced_index(self):
        """
        Get the last index which has been synced from the server.

        Will also initialize the array if this hasn't happened already.
        TODO: seems hacky to init_data here.

        Returns:
            int: the last flat index which has been synced from the server,
                or -1 if no data has been synced.
        """
        if not hasattr(self, 'synced_index'):
            self.init_data()
            self.synced_index = -1

        return self.synced_index

    def get_changes(self, synced_index):
        """
        Find changes since the last sync of this array.

        Args:
            synced_index (int): The last flat index which has already
                been synced.

        Returns:
            Union[dict, None]: None if there is no new data. If there is,
                returns a dict with keys:
                    start (int): the flat index of the first returned value.
                    stop (int): the flat index of the last returned value.
                    vals (List[float]): the new values
        """
        latest_index = self.last_saved_index
        if latest_index is None:
            latest_index = -1
        if self.modified_range:
            latest_index = max(latest_index, self.modified_range[1])

        vals = [
            self.ndarray[np.unravel_index(i, self.ndarray.shape)]
            for i in range(synced_index + 1, latest_index + 1)
        ]

        if vals:
            return {
                'start': synced_index + 1,
                'stop': latest_index,
                'vals': vals
            }

    def apply_changes(self, start, stop, vals):
        """
        Insert new synced values into the array.

        To be be called in a ``PULL_FROM_SERVER`` ``DataSet`` using results
        returned by ``get_changes`` from the ``DataServer``.

        TODO: check that vals has the right length?

        Args:
            start (int): the flat index of the first new value.
            stop (int): the flat index of the last new value.
            vals (List[float]): the new values
        """
        for i, val in enumerate(vals):
            index = np.unravel_index(i + start, self.ndarray.shape)
            self.ndarray[index] = val
        self.synced_index = stop

    def __repr__(self):
        array_id_or_none = f' {self.array_id}' if self.array_id else ''
        return '{}[{}]:{}\n{}'.format(self.__class__.__name__,
                                      ','.join(map(str, self.shape)),
                                      array_id_or_none, repr(self.ndarray))

    def snapshot(self, update=False):
        """JSON representation of this DataArray."""
        snap = {'__class__': full_class(self)}

        snap.update(self._snapshot_input)

        for attr in self.SNAP_ATTRS:
            snap[attr] = getattr(self, attr)

        return snap

    def fraction_complete(self):
        """
        Get the fraction of this array which has data in it.

        Or more specifically, the fraction of the latest point in the array
        where we have touched it.

        Returns:
            float: fraction of array which is complete, from 0.0 to 1.0
        """
        if self.ndarray is None:
            return 0.0

        last_index = -1
        if self.last_saved_index is not None:
            last_index = max(last_index, self.last_saved_index)
        if self.modified_range is not None:
            last_index = max(last_index, self.modified_range[1])
        if getattr(self, 'synced_index', None) is not None:
            last_index = max(last_index, self.synced_index)

        return (last_index + 1) / self.ndarray.size

    @property
    def units(self):
        warn_units('DataArray', self)
        return self.unit

    def to_xarray(self) -> xr.DataArray:
        """ Return this DataArray as an xarray dataarray

        Returns:
            DataArray in xarray format
        """
        xarray_dictionary = data_array_to_xarray_dictionary(self)
        xarray_dataarray = xr.DataArray.from_dict(xarray_dictionary)
        return xarray_dataarray

    @classmethod
    def from_xarray(cls, xarray_dataarray: xr.DataArray, array_id: Optional[str] = None) -> 'DataArray':
        """ Create a DataArray from an xarray DataArray

        Args:
            array_id: Array id for the new DataArray. If None, then use the first data variable from the argument
        Returns:
            Created xarray DataArray
        """
        xarray_dict = xarray_dataarray.to_dict()
        if array_id is None:
            array_id = list(xarray_dict['dims'])[0]
        data_array = xarray_data_array_dictionary_to_data_array(array_id, xarray_dict, is_setpoint=False)
        return data_array


def data_array_to_xarray_dictionary(data_array: DataArray) -> Dict[str, Any]:
    """Convert DataArray to a dictionary in xarray format.

    Args:
        data_array: The DataArray to convert.

    Returns:
        dict: A dictionary containing the data in xarray format.
    """
    key_mapping = {"unit": "unit", "name": "name", "label": "label"}

    data_dictionary = {
        target_key: getattr(data_array, key) for key, target_key in key_mapping.items()
    }
    data_dictionary['long_name'] = data_array.name
    if data_array.is_setpoint:
        data_dictionary["dims"] = tuple([data_array.array_id])
        data_dictionary["depends_on"] = data_dictionary["dims"]
        data = data_array.ndarray
        # flatten data, assumes setpoint is uniform as for a normal gridded dataset
        while len(data.shape) > 1:
            data = data[0, ..., :]
        data_dictionary["data"] = data
    else:
        if data_array.set_arrays:
            data_dictionary["dims"] = tuple(a.array_id for a in data_array.set_arrays)
            data_dictionary["depends_on"] = data_dictionary["dims"]
        data_dictionary["data"] = data_array.ndarray

    return data_dictionary


def xarray_data_array_dictionary_to_data_array(
        array_id: str, array_dictionary: Dict[str, Any], is_setpoint: bool = False, preset_data=None):
    """Convert xarray dictionary to a DataArray

    This conversion is for bith the data array and the the internal xarray structure, e.g. the datavars and coords.
    Args:
        array_id: Create the new DataArray with this id
        array_dictionary: Data to convert
        is_setpoint: Passed to the DataArray constructor
        preset_data: If None use the data from the dictionary, otherwise use the specified data.

    Returns:
        dict: A dictionary containing the data in xarray format.
    """
    if preset_data is None:
        preset_data = np.array(array_dictionary["data"])
    array_name = array_dictionary.get("name", array_id)

    array_full_name = array_dictionary.get("long_name", array_name)
    data_array = DataArray(
        name=array_name,
        full_name=array_full_name,
        label=array_dictionary.get("label", ""),
        unit=array_dictionary.get("unit", None),
        is_setpoint=is_setpoint,
        shape=preset_data.shape,
        array_id=array_id,
        preset_data=preset_data,
    )
    return data_array



from keyword import iskeyword
from numbers import Number

import hypothesis.strategies as hst
import pytest
from hypothesis import assume, given
from numpy import ndarray

from qcodes.dataset.descriptions.param_spec import ParamSpec, ParamSpecBase


def valid_identifier(**kwargs):
    """Return a strategy which generates a valid Python Identifier"""
    if 'min_size' not in kwargs:
        kwargs['min_size'] = 4
    return hst.text(
        alphabet="abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789_",
        **kwargs).filter(
        lambda x: x[0].isalpha() and x.isidentifier() and not (iskeyword(x))
    )


# This strategy generates a dict of kwargs needed to instantiate a valid
# ParamSpec object
valid_paramspec_kwargs = hst.fixed_dictionaries(
    {'name': valid_identifier(min_size=1, max_size=6),
     'paramtype': hst.sampled_from(['numeric', 'array', 'text']),
     'label': hst.one_of(hst.none(), hst.text(min_size=0, max_size=6)),
     'unit': hst.one_of(hst.none(), hst.text(min_size=0, max_size=2)),
     'depends_on': hst.lists(hst.text(min_size=1, max_size=3),
                             min_size=0, max_size=3),
     'inferred_from': hst.lists(hst.text(min_size=1, max_size=3),
                                min_size=0, max_size=3)
     })


@pytest.fixture
def version_0_dicts():
    sers = []
    sers.append({'name': 'dmm_v1',
                 'paramtype': 'numeric',
                 'label': 'Gate v1',
                 'unit': 'V',
                 'inferred_from': [],
                 'depends_on': ['dac_ch1', 'dac_ch2']})
    sers.append({'name': 'some_name',
                 'paramtype': 'array',
                 'label': 'My Array ParamSpec',
                 'unit': 'Ars',
                 'inferred_from': ['p1', 'p2'],
                 'depends_on': []})
    return sers


@pytest.fixture
def version_0_objects():
    """
    The ParamSpecs that the dictionaries above represent
    """
    ps = []
    ps.append(ParamSpec('dmm_v1', paramtype='numeric', label='Gate v1',
                        unit='V', inferred_from=[],
                        depends_on=['dac_ch1', 'dac_ch2']))
    ps.append(ParamSpec('some_name', paramtype='array',
                        label='My Array ParamSpec', unit='Ars',
                        inferred_from=['p1', 'p2'], depends_on=[]))
    return ps


@given(name=hst.text(min_size=1),
       sp1=hst.text(min_size=1), sp2=hst.text(min_size=1),
       inff1=hst.text(min_size=1), inff2=hst.text(min_size=1),
       paramtype=hst.lists(
           elements=hst.sampled_from(['numeric', 'array', 'text']),
           min_size=6, max_size=6))
def test_creation(name, sp1, sp2, inff1, inff2, paramtype):
    invalid_types = ['np.array', 'ndarray', 'lala', '', Number,
                     ndarray, 0, None]
    for inv_type in invalid_types:
        with pytest.raises(ValueError):
            ParamSpec(name, inv_type)

    if not inff1.isidentifier():
        inff1 = 'inff1'

    if not sp1.isidentifier():
        sp1 = 'sp1'

    if not name.isidentifier():
        with pytest.raises(ValueError):
            ps = ParamSpec(name, paramtype[0], label=None, unit='V',
                           inferred_from=(inff1, inff2),
                           depends_on=(sp1, sp2))
        name = 'name'

    ps = ParamSpec(name, paramtype[1], label=None, unit='V',
                   inferred_from=(inff1, inff2),
                   depends_on=(sp1, sp2))

    assert ps.inferred_from == f'{inff1}, {inff2}'
    assert ps.depends_on == f'{sp1}, {sp2}'

    ps1 = ParamSpec(sp1, paramtype[2])
    p1 = ParamSpec(name, paramtype[3], depends_on=(ps1, sp2))
    assert p1.depends_on == ps.depends_on

    ps2 = ParamSpec(inff1, paramtype[4])
    p2 = ParamSpec(name, paramtype[5], inferred_from=(ps2, inff2))
    assert p2.inferred_from == ps.inferred_from


@given(name=hst.text(min_size=1))
def test_repr(name):
    okay_types = ['array', 'numeric', 'text']

    for okt in okay_types:
        if name.isidentifier():
            ps = ParamSpec(name, okt)
            expected_repr = (f"ParamSpec('{name}', '{okt}', '', '', "
                             "inferred_from=[], depends_on=[])")
            assert ps.__repr__() == expected_repr
        else:
            with pytest.raises(ValueError):
                ps = ParamSpec(name, okt)


alphabet = "".join(chr(i) for i in range(ord("a"), ord("z")))


@given(
    name1=hst.text(min_size=4, max_size=100, alphabet=alphabet),
    name2=hst.text(min_size=4, max_size=100, alphabet=alphabet),
    name3=hst.text(min_size=4, max_size=100, alphabet=alphabet)
)
def test_depends_on(name1, name2, name3):
    ps2 = ParamSpec(name2, "numeric")
    ps3 = ParamSpec(name3, "numeric")

    ps1 = ParamSpec(name1, "numeric", depends_on=[ps2, ps3, 'foo'])

    assert ps1.depends_on == f"{ps2.name}, {ps3.name}, foo"
    assert ps1.depends_on_ == [ps2.name, ps3.name, "foo"]

    with pytest.raises(ValueError,
                       match=f"ParamSpec {name1} got string foo as depends_on. "
                       "It needs a Sequence of ParamSpecs or strings"):
        ParamSpec(name1, "numeric", depends_on='foo')


@given(
    name1=hst.text(min_size=4, max_size=100, alphabet=alphabet),
    name2=hst.text(min_size=4, max_size=100, alphabet=alphabet),
    name3=hst.text(min_size=4, max_size=100, alphabet=alphabet)
)
def test_inferred_from(name1, name2, name3):
    ps2 = ParamSpec(name2, "numeric")
    ps3 = ParamSpec(name3, "numeric")

    ps1 = ParamSpec(name1, "numeric", inferred_from=[ps2, ps3, 'bar'])

    assert ps1.inferred_from == f"{ps2.name}, {ps3.name}, bar"
    assert ps1.inferred_from_ == [ps2.name, ps3.name, "bar"]

    with pytest.raises(ValueError,
                       match=f"ParamSpec {name1} got string foo as "
                       f"inferred_from. "
                       "It needs a Sequence of ParamSpecs or strings"):
        ParamSpec(name1, "numeric", inferred_from='foo')


@given(
    name1=hst.text(min_size=4, max_size=100, alphabet=alphabet),
    name2=hst.text(min_size=4, max_size=100, alphabet=alphabet)
)
def test_copy(name1, name2):
    ps_indep = ParamSpec(name1, "numeric")
    ps = ParamSpec(name2, "numeric", depends_on=[ps_indep, 'other_param'])
    ps_copy = ps.copy()

    assert ps_copy == ps
    assert hash(ps_copy) == hash(ps)

    att_names = ["name", "type", "label", "unit",
                 "_inferred_from", "_depends_on"]

    attributes = {}
    for att in att_names:
        val = getattr(ps, att)
        valc = getattr(ps_copy, att)
        assert val == valc
        attributes[att] = val

    # Modifying the copy should not change the original
    for att in att_names:
        if not att.startswith('_'):
            setattr(ps_copy, att, attributes[att] + "_modified")
        else:
            setattr(ps_copy, att, attributes[att] + ['bob'])
        assert getattr(ps, att) == attributes[att]

    assert ps_copy != ps
    assert hash(ps_copy) != hash(ps)


def test_convert_to_dict():
    p1 = ParamSpec('p1', 'numeric', 'paramspec one', 'no unit',
                   depends_on=['some', 'thing'], inferred_from=['bab', 'bob'])

    ser = p1._to_dict()

    assert ser['name'] == p1.name
    assert ser['paramtype'] == p1.type
    assert ser['label'] == p1.label
    assert ser['unit'] == p1.unit
    assert ser['depends_on'] == p1._depends_on
    assert ser['inferred_from'] == p1._inferred_from


def test_from_dict(version_0_dicts, version_0_objects):
    for sdict, ps in zip(version_0_dicts, version_0_objects):
        deps = ParamSpec._from_dict(sdict)
        assert ps == deps


@given(paramspecs=hst.lists(valid_paramspec_kwargs, min_size=2, max_size=2))
def test_hash(paramspecs):
    p1 = ParamSpec(**paramspecs[0])
    p2 = ParamSpec(**paramspecs[1])

    # call __hash__
    p1_h = hash(p1)
    p2_h = hash(p2)

    # make a set
    p_set = {p1, p2}

    # test that the hash equality follows object equality
    if p1 == p2:
        assert p1_h == p2_h
        assert 1 == len(p_set)
    else:
        assert p1_h != p2_h
        assert 2 == len(p_set)


@given(paramspecs=hst.lists(valid_paramspec_kwargs, min_size=6, max_size=6),
       add_to_1_inf=hst.booleans(),
       add_to_1_dep=hst.booleans(),
       add_to_2_inf=hst.booleans(),
       add_to_2_dep=hst.booleans(),
       )
def test_hash_with_deferred_and_inferred_as_paramspecs(
        paramspecs, add_to_1_inf, add_to_1_dep, add_to_2_inf, add_to_2_dep):
    """
    Test that hashing works if 'inferred_from' and/or 'depends_on' contain
    actual ParamSpec instances and not just strings.
    """
    assume(add_to_1_inf or add_to_1_dep or add_to_2_inf or add_to_2_dep)

    # Add ParamSpecs to 'inferred_from' and/or 'depend_on' lists next to
    # strings (that are generated by the main strategy)
    if add_to_1_inf:
        paramspecs[0]['inferred_from'].append(ParamSpec(**paramspecs[2]))
    if add_to_1_dep:
        paramspecs[0]['depends_on'].append(ParamSpec(**paramspecs[3]))
    if add_to_2_inf:
        paramspecs[1]['inferred_from'].append(ParamSpec(**paramspecs[4]))
    if add_to_2_dep:
        paramspecs[1]['depends_on'].append(ParamSpec(**paramspecs[5]))

    p1 = ParamSpec(**paramspecs[0])
    p2 = ParamSpec(**paramspecs[1])

    # call __hash__
    p1_h = hash(p1)
    p2_h = hash(p2)

    # make a set
    p_set = {p1, p2}

    # test that the hash equality follows object equality
    if p1 == p2:
        assert p1_h == p2_h
        assert 1 == len(p_set)
    else:
        assert p1_h != p2_h
        assert 2 == len(p_set)


@given(paramspecs=hst.lists(valid_paramspec_kwargs, min_size=1, max_size=1))
def test_base_version(paramspecs):

    kwargs = paramspecs[0]

    ps = ParamSpec(**kwargs)
    ps_base = ParamSpecBase(name=kwargs['name'],
                            paramtype=kwargs['paramtype'],
                            label=kwargs['label'],
                            unit=kwargs['unit'])

    assert ps.base_version() == ps_base


def test_not_eq_for_list_attr():
    """
    test that two paramspecs that differ only
    in list attrs are different
    """

    p1 = ParamSpec(name='foo',
                   paramtype='numeric',
                   depends_on=['a', 'b'])
    p2 = ParamSpec(name='foo',
                   paramtype='numeric',
                   depends_on=['c', 'd'])
    assert p1 != p2


def test_not_eq_for_str_attr():
    """
    test that two paramspecs that differ only
    in str attrs are different
    """

    p1 = ParamSpec(name='foo',
                   label='myfoo',
                   paramtype='numeric',
                   depends_on=['a', 'b'])
    p2 = ParamSpec(name='foo',
                   label='someotherfoo',
                   paramtype='numeric',
                   depends_on=['a', 'b'])
    assert p1 != p2


def test_not_eq_non_paramspec():
    """
    test that two paramspecs that differ only
    in str attrs are different
    """

    p1 = ParamSpec(name='foo',
                   label='myfoo',
                   paramtype='numeric',
                   depends_on=['a', 'b'])
    p2 = 1
    assert p1 != p2



from enum import Enum, IntEnum, IntFlag
from typing import Union, Sequence


class StrEnum(str, Enum):
    pass

class ChannelName(StrEnum):
    A = 'CH1'
    B = 'CH2'
    C = 'CH3'
    D = 'CH4'
    E = 'CH5'
    F = 'CH6'
    G = 'CH7'
    H = 'CH8'
    I = 'CH9'
    J = 'CH10'
    Z = 'XDATA'


class MeasurementError(IntEnum):
    """
    The list below enumerates all possible compliance errors and assigns
    integer 0 or 1 to them. Assigning integer 1 means the data is compliant
    and assigning 0 means it not compliant. As seen from the list, all cases
    except when the output is N are non-compliant. To know the reason for
    Non compliance in the individual case one may refer to
    :class:`.MeasurementStatus`
    """
    C = 0
    N = 1
    T = 0
    V = 0
    X = 0
    U = 0
    D = 0
    G = 0
    S = 0


class MeasurementStatus(StrEnum):
    """
    Contains the meanings of possible compliance errors. One may look at
    this list to figure out the reason for the non-compliant data.
    """
    C = 'Reached compliance limit.'
    N = 'No status error occurred.'
    T = 'Another channel reached compliance limit.'
    V = 'Measurement data is over the measurement range.' \
        ' Or the sweep measurement was aborted by the automatic' \
        ' stop function or power compliance.' \
        ' D will be 199.999E+99 (no meaning).'
    X = 'One or more channels are oscillating. ' \
        'Or source output did not settle before measurement.'
    U = 'CMU is in the NULL loop unbalance condition.'
    D = 'CMU is in the IV amplifier saturation condition.'
    G = 'For linear or binary search measurement, ' \
        'the target value was not found within the search range.' \
        ' Returns the source output value. ' \
        'For quasi-pulsed spot measurement, ' \
        'the detection time was over the limit ' \
        '(3 s for Short mode, 12 s for Long mode).'
    S = 'For linear or binary search measurement, ' \
        'the search measurement was stopped. ' \
        'Returns the source output value. ' \
        'See status of Data_sense.' \
        'For quasi-pulsed spot measurement, ' \
        'output slew rate was too slow to perform the settling detection.' \
        'Or quasi-pulsed source channel reached the current compliance' \
        ' before the source output voltage changed 10 V from the start voltage.'


class ModuleKind(StrEnum):
    """
    This enumeration is internal to this driver, and is meant to classify
    difference B1500 modules, as well as provide a default naming scheme for
    them (hence the `StrEnum`).
    """
    SMU = 'SMU'  # Source/Monitor Unit
    CMU = 'CMU'  # Capacitance Measurement Unit
    WGFMU = 'WGFMU'  # Waveform Generator/Fast Measurement Unit


class SlotNr(IntEnum):
    ALL = 0
    MAINFRAME = 11
    SLOT01 = 1
    SLOT02 = 2
    SLOT03 = 3
    SLOT04 = 4
    SLOT05 = 5
    SLOT06 = 6
    SLOT07 = 7
    SLOT08 = 8
    SLOT09 = 9
    SLOT10 = 10


class ChNr(IntEnum):
    SLOT_01_CH1 = 1
    SLOT_02_CH1 = 2
    SLOT_03_CH1 = 3
    SLOT_04_CH1 = 4
    SLOT_05_CH1 = 5
    SLOT_06_CH1 = 6
    SLOT_07_CH1 = 7
    SLOT_08_CH1 = 8
    SLOT_09_CH1 = 9
    SLOT_10_CH1 = 10

    SLOT_01_CH2 = 102
    SLOT_02_CH2 = 202
    SLOT_03_CH2 = 302
    SLOT_04_CH2 = 402
    SLOT_05_CH2 = 502
    SLOT_06_CH2 = 602
    SLOT_07_CH2 = 702
    SLOT_08_CH2 = 802
    SLOT_09_CH2 = 902
    SLOT_10_CH2 = 1002


ChannelList = Sequence[Union[ChNr, int]]


class Abort(IntEnum):
    DISABLED = 1
    ENABLED = 2


class TriggerPort(IntEnum):
    EXT_TRIG_IN = -1
    EXT_TRIG_OUT = -2
    DIO_1 = 1
    DIO_2 = 2
    DIO_3 = 3
    DIO_4 = 4
    DIO_5 = 5
    DIO_6 = 6
    DIO_7 = 7
    DIO_8 = 8
    DIO_9 = 9
    DIO_10 = 10
    DIO_11 = 11
    DIO_12 = 12
    DIO_13 = 13
    DIO_14 = 14
    DIO_15 = 15
    DIO_16 = 16


class SweepMode(IntEnum):
    LINEAR = 1
    LOG = 2
    LINEAR_TWO_WAY = 3
    LOG_TWO_WAY = 4


class LinearSweepMode(IntEnum):
    LINEAR = 1
    LINEAR_TWO_WAY = 3


class APIVersion(IntEnum):
    B1500 = 0
    CLASSIC = 1


class VOutputRange(IntEnum):
    AUTO = 0

    MIN_0V2 = 2
    MIN_0V5 = 5
    MIN_2V = 20
    MIN_5V = 50
    MIN_20V = 200
    MIN_40V = 400
    MIN_100V = 1000
    MIN_200V = 2000
    MIN_500V = 5000
    MIN_1500V = 15000
    MIN_3000V = 30000
    MIN_10000V = 103


class VMeasRange(IntEnum):
    AUTO = 0

    MIN_0V2 = 2
    MIN_0V5 = 5
    MIN_2V = 20
    MIN_5V = 50
    MIN_20V = 200
    MIN_40V = 400
    MIN_100V = 1000
    MIN_200V = 2000
    MIN_500V = 5000
    MIN_1500V = 15000
    MIN_3000V = 30000
    MIN_10000V = 103

    FIX_0V2 = -2
    FIX_0V5 = -5
    FIX_2V = -20
    FIX_5V = -50
    FIX_20V = -200
    FIX_40V = -400
    FIX_100V = -1000
    FIX_200V = -2000
    FIX_500V = -5000
    FIX_1500V = -15000
    FIX_3000V = -30000
    FIX_10000V = -103


class IOutputRange(IntEnum):
    AUTO = 0

    MIN_1pA = 8
    MIN_10pA = 9
    MIN_100pA = 10
    MIN_1nA = 11
    MIN_10nA = 12
    MIN_100nA = 13
    MIN_1uA = 14
    MIN_10uA = 15
    MIN_100uA = 16
    MIN_1mA = 17
    MIN_10mA = 18
    MIN_100mA = 19
    MIN_1A = 20
    MIN_2A = 21
    MIN_20A = 22
    MIN_40A = 23

    FIX_500A = 26
    FIX_2000A = 28


class IMeasRange(IntEnum):
    AUTO = 0

    MIN_1pA = 8
    MIN_10pA = 9
    MIN_100pA = 10
    MIN_1nA = 11
    MIN_10nA = 12
    MIN_100nA = 13
    MIN_1uA = 14
    MIN_10uA = 15
    MIN_100uA = 16
    MIN_1mA = 17
    MIN_10mA = 18
    MIN_100mA = 19
    MIN_1A = 20
    MIN_2A = 21
    MIN_20A = 22
    MIN_40A = 23

    MIN_500A = 26
    MIN_2000A = 28

    FIX_1pA = -8
    FIX_10pA = -9
    FIX_100pA = -10
    FIX_1nA = -11
    FIX_10nA = -12
    FIX_100nA = -13
    FIX_1uA = -14
    FIX_10uA = -15
    FIX_100uA = -16
    FIX_1mA = -17
    FIX_10mA = -18
    FIX_100mA = -19
    FIX_1A = -20
    FIX_2A = -21
    FIX_20A = -22
    FIX_40A = -23

    FIX_500A = -26
    FIX_2000A = -28


OutputRange = Union[VOutputRange, IOutputRange]

MeasureRange = Union[VMeasRange, IMeasRange]


class RangingMode(IntEnum):
    AUTO = 0
    FIXED = 2


class CompliancePolarityMode(IntEnum):
    AUTO = 0
    MANUAL = 1


class LinearSearchMode(IntEnum):
    VALUE_LEQ_TARGET = 0
    VALUE_GEQ_TARGET = 1


class BinarySearchMode(IntEnum):
    LIMIT = 0
    REPEAT_COUNT = 1


class Polarity(IntEnum):
    NEGATIVE = 0
    POSITIVE = 1


class CalibrationType(IntEnum):
    OPEN = 1
    SHORT = 2
    LOAD = 3


class AutoPeriod(IntEnum):
    AUTO_EFFECTIVE_MINIMUM = 0
    AUTO_LONGEST = -1


#: Command specific Enums

class AAD:
    class Type(IntEnum):
        HIGH_SPEED = 0
        HIGH_RESOLUTION = 1
        HIGH_SPEED_PULSED = 2


class ACT:
    class Mode(IntEnum):
        AUTO = 0
        PLC = 2


class ADJ:
    class Mode(IntEnum):
        AUTO = 0
        MANUAL = 1
        LOAD_ADAPTIVE = 2


class ADJQuery:
    class Mode(IntEnum):
        USE_LAST = 0
        MEASURE = 1
    class Response(IntEnum):
        PASSED = 0
        FAILED = 1
        ABORTED = 2
        NOT_PERFORMED = 3


class AIT:
    class Type(IntEnum):
        HIGH_SPEED = 0
        HIGH_RESOLUTION = 1
        HIGH_SPEED_PULSED = 2

    class Mode(IntEnum):
        AUTO = 0
        MANUAL = 1
        NPLC = 2
        MEAS_TIME_MODE = 3


class AV:
    class Mode(IntEnum):
        AUTO = 0
        MANUAL = 1


class BDM:
    class Interval(IntEnum):
        SHORT = 0
        LONG = 1

    class Mode(IntEnum):
        VOLTAGE = 0
        CURRENT = 1


class BSM:
    class Mode(IntEnum):
        NORMAL = 0
        CAUTIOUS = 1

    class Post(IntEnum):
        START_VAL = 1
        STOP_VAL = 2
        OUTPUT_AT_SEARCH_TARGET = 3


class BSVM:
    class DataOutputMode(IntEnum):
        SEARCH = 0
        SEARCH_AND_SENSE = 1


class CALResponse(IntFlag):
    PASSED = 0
    SLOT_1_FAILED = 1
    SLOT_2_FAILED = 2
    SLOT_3_FAILED = 4
    SLOT_4_FAILED = 8
    SLOT_5_FAILED = 16
    SLOT_6_FAILED = 32
    SLOT_7_FAILED = 64
    SLOT_8_FAILED = 128
    SLOT_9_FAILED = 256
    SLOT_10_FAILED = 512
    MAINFRAME_FAILED = 1024


class CORR:
    class Response(IntEnum):
        SUCCESSFUL = 0
        FAILED = 1
        ABORTED = 2


class CORRST:
    class Response(IntEnum):
        OFF = 0
        ON = 1


class CLCORR:
    class Mode(IntEnum):
        CLEAR_ONLY = 1
        CLEAR_AND_SET_DEFAULT_FREQ = 2


class CMM:
    class Mode(IntEnum):
        COMPLIANCE_SIDE = 0
        FORCE_SIDE = 3
        CURRENT = 1
        VOLTAGE = 2
        CURRENT_AND_VOLTAGE = 4


class DCORR:
    class Mode(IntEnum):
        Cp_G = 100
        Ls_Rs = 400


class DIAG:
    class Item(IntEnum):
        TRIGGER_IO = 1
        HIGH_VOLTAGE_LED = 3
        DIGITAL_IO = 4
        INTERLOCK_OPEN = 6
        INTERLOCK_CLOSE = 7


class ERCMAGRD:
    class Guard(IntEnum):
        COMMON_AC = 1
        FLOATING = 2


class ERHPP:
    class Path(IntEnum):
        OPEN = 0
        HVSMU = 1
        HCSMU = 2
        HPSMU = 3
        HVSMU_WITH_R_SERIES = 4


class ERHPQG:
    class State(IntEnum):
        OPEN = 0
        GATE_CHARGE = 1
        IV = 2


class ERHVP:
    class State(IntEnum):
        OPEN = 0
        HVSMU = 1
        HVMCU = 2


class ERHVPV:
    class State(IntEnum):
        OPEN = 0
        HVMCU_DC = 1
        CAPACITANCE_CHARGE = 2


class ERMOD:
    class Mode(IntEnum):
        GENERAL_PURPOSE = 0
        SMU_PGU_SELECTOR_16440A = 1
        N1258A_N1259A = 2
        N1265A = 4
        N1266A = 8
        N1268A = 16
        N1272A = 32


class ERPFDP:
    class State(IntEnum):
        OPEN = 0
        UHCUL_AND_UHCUH = 1
        GNDU_AND_HVSMU_HVMCU = 2
        GNDU_AND_MP_HPSMU = 3
        GNDU_AND_OPEN = 4


class ERPFGP:
    class State(IntEnum):
        OPEN = 0
        CONNECTED = 1


class ERPFGR:
    class State:
        R_0 = 0
        R_10 = 10
        R_100 = 100
        R_1000 = 1000


class ERR:
    class Mode(IntEnum):
        READ_ALL = 0
        READ_ONE = 1


class ERRX:
    class Mode(IntEnum):
        CODE_AND_MESSAGE = 0
        CODE_ONLY = 1


class ERSSP:
    class Port(IntEnum):
        SELECTOR_1_OUT_1 = 0
        SELECTOR_1_OUT_2 = 1
        SELECTOR_2_OUT_1 = 2
        SELECTOR_2_OUT_2 = 3

    class Status(IntEnum):
        ALL_OPEN = 0
        SMU_ON = 1
        PGU_ON = 2
        PGU_OPEN = 3


class FMT:
    class Format(IntEnum):
        ASCII_12_DIGITS_WITH_HEADER_CRLF_EOI = 1
        ASCII_12_DIGITS_NO_HEADER_CRLF_EOI = 2
        BINARY_4_BYTE_CRLF_EOI = 3
        BINARY_4_BYTE_EOI = 4
        ASCII_12_DIGITS_WITH_HEADER_COMMA = 5
        ASCII_13_DIGITS_WITH_HEADER_CRLF_EOI = 11
        ASCII_13_DIGITS_NO_HEADER_CRLF_EOI = 12
        BINARY_8_BYTE_CRLF_EOI = 13
        BINARY_8_BYTE_EOI = 14
        ASCII_13_DIGITS_WITH_HEADER_COMMA = 15
        ASCII_13_DIGITS_WITH_HEADER_CRLF_EOI_4155_4156_COMPATIBLE = 21
        ASCII_13_DIGITS_NO_HEADER_CRLF_EOI_4155_4156_COMPATIBLE = 22
        ASCII_13_DIGITS_WITH_HEADER_COMMA_4155_4156_COMPATIBLE = 25

    class Mode(IntEnum):
        ONLY_MEASUREMENT_DATA = 0
        PRIMARY_SOURCE_OUTPUT_DATA = 1
        SYNCHRONOUS_SWEEP_SOURCE_OUTPUT_DATA = 2
        SWEEP_SOURCE_1 = 1
        SWEEP_SOURCE_2 = 2
        SWEEP_SOURCE_3 = 3
        SWEEP_SOURCE_4 = 4
        SWEEP_SOURCE_5 = 5
        SWEEP_SOURCE_6 = 6
        SWEEP_SOURCE_7 = 7
        SWEEP_SOURCE_8 = 8
        SWEEP_SOURCE_9 = 9
        SWEEP_SOURCE_10 = 10


class HVSMUOP:
    class SourceRange(IntEnum):
        SINGLE_CH = 1
        PLUSMINUS_1500V = 2
        PLUS_3000V = 3
        MINUS_3000V = 4


class IMP:
    class MeasurementMode(IntEnum):
        R_X = 1
        G_X = 2
        Z_THETA_RAD = 10
        Z_THETA_DEG = 11
        Y_THETA_RAD = 20
        Y_THETA_DEG = 21
        Cp_G = 100
        Cp_D = 101
        Cp_Q = 102
        Cp_Rp = 103
        Cs_Rs = 200
        Cs_D = 201
        Cs_Q = 202
        Lp_G = 300
        Lp_D = 301
        Lp_Q = 302
        Lp_Rp = 303
        Ls_Rs = 400
        Ls_D = 401
        Ls_Q = 402

    class Name(StrEnum):
        R = 'Resistance'
        G = 'Conductance'
        Z = 'Impedance'
        Y = 'Admittance'
        Cp = 'Parallel Capacitance'
        Cs = 'Series Capacitance'
        Lp = 'Parallel Inductance'
        Ls = 'Series Inductance'
        X = 'Reactance'
        B = 'Susceptance'
        THETA_RAD = 'Phase'
        THETA_DEG = 'Phase'
        D = 'Dissipation Factor'
        Q = 'Quality Factor'
        Rp = 'Parallel Resistance'
        Rs = 'Series Resistance'

    class Unit(StrEnum):
        R = 'ohms'
        G = 'S'
        Z = 'ohms'
        Y = 'S'
        Cp = 'F'
        Cs = 'F'
        Lp = 'H'
        Ls = 'H'
        X = 'ohms'
        B = 'S'
        THETA_RAD = 'radian'
        THETA_DEG = 'degree'
        D = ''
        Q = ''
        Rp = 'ohms'
        Rs = 'ohms'


class LIM:
    class Mode(IntEnum):
        VOLTAGE = 1
        CURRENT = 2


class LRN:
    class Type(IntEnum):
        OUTPUT_SWITCH = 0

        SLOT1_STATUS = 1
        SLOT2_STATUS = 2
        SLOT3_STATUS = 3
        SLOT4_STATUS = 4
        SLOT5_STATUS = 5
        SLOT6_STATUS = 6
        SLOT7_STATUS = 7
        SLOT8_STATUS = 8
        SLOT9_STATUS = 9
        SLOT10_STATUS = 10

        FILTER = 30

        TM_AV_CM_FMT_MM_SETTINGS = 31

        MEASUREMENT_RANGING_STATUS = 32

        STAIRCASE_SWEEP_MEASUREMENT_SETTINGS = 33

        PULSED_SOURCE_SETTINGS = 34

        QUASI_PULSED_SOURCE_SETTINGS = 37

        DIGITAL_IO_SETTINGS = 38

        CHANNEL_MAPPING = 40

        SMU_MEASUREMENT_OPERATION = 46

        SAMPLING_MEASUREMENT_SETTINGS = 47

        QUASI_STATIC_CV_MEASUREMENT_SETTINGS = 49

        LINEAR_SEARCH_MEASUREMENT_SETTINGS = 50

        BINARY_SEARCH_MEASUREMENT_SETTINGS = 51

        SMU_RESISTOR_STATUS = 53

        AUTOR_RANGING_MODE = 54

        ADC_SETTINGS = 55

        ADC_AVERAGING_INTEGRATION_TIME_SETTINGS = 56

        SOURCE_MEASURE_WAIT_TIME = 57

        TRIGGER_SETTINGS = 58

        MULTI_CHANNEL_SWEEP_SOURCE_SETTINGS = 59

        TIMESTAMP_SETTING = 60

        DISPLAY_SETTING = 61

        ASU_CONNECTION_PATH = 62

        PICOAMPERE_RANGING_MODE = 63

        ASU_CONNECTION_STATUS_INDICATOR = 64

        MFCMU_MEASUREMENT_MODE = 70

        MFCMU_DATA_OUTPUT_MODE = 71

        MFCMU_ADC_SETTING = 72

        MFCMU_MEASUREMENT_RANGE = 73

        SCUU_CONNECTION_STATUS_INDICATOR = 80

        SCUU_CONNECTION_PATH = 81

        MFCMU_ADJUSTMENT_MODE = 90

        CV_DC_BIAS_SWEEP_MEASUREMENT_SETTINGS = 100

        PULSED_SPOT_C_CV_MEASUREMENT_SETTINGS = 101

        C_F_SWEEP_MEASUREMENT_SETTINGS = 102

        CV_AC_LEVEL_SWEEP_MEASUREMENT_SETTINGS = 103

        C_T_SAMPLING_MEASUREMENT_SETTINGS = 104

        MULTI_CHANNEL_PULSED_SPOT_MEASUREMENT_SETTINGS = 105

        MULTI_CHANNEL_PULSED_SWEEP_MEASUREMENT_SETTINGS = 106

        PARALLEL_MEASUREMENT_MODE_SETTING = 110


class LSM:
    class Post(IntEnum):
        START_VAL = 1
        STOP_VAL = 2
        OUTPUT_AT_SEARCH_TARGET = 3


class LSVM(BSVM):
    pass


class MCPNX:
    class Mode(IntEnum):
        VOLTAGE = 1
        CURRENT = 2


class MCPWNX:
    class Mode(IntEnum):
        VOLTAGE = 1
        CURRENT = 2


class ML:
    class Mode(IntEnum):
        LINEAR = 1
        LOG_10_PER_DECADE = 2
        LOG_25_PER_DECADE = 3
        LOG_50_PER_DECADE = 4
        LOG_100_PER_DECADE = 5
        LOG_250_PER_DECADE = 6
        LOG_500_PER_DECADE = 7


class MM:
    class Mode(IntEnum):
        SPOT = 1
        STAIRCASE_SWEEP = 2
        PULSED_SPOT = 3
        PULSED_SWEEP = 4
        STAIRCASE_SWEEP_WITH_PULSED_BIAS = 5
        QUASI_PULSED_SPOT = 9
        SAMPLING = 10
        QUASI_STATIC_CV = 13
        LINEAR_SEARCH = 14
        BINARY_SEARCH = 15
        MULTI_CHANNEL_SWEEP = 16
        SPOT_C = 17
        CV_DC_SWEEP = 18
        PULSED_SPOT_C = 19
        PULSED_SWEEP_CV = 20
        CF_SWEEP = 22
        CV_AC_SWEEP = 23
        CT_SAMPLING = 26
        MULTI_CHANNEL_PULSED_SPOT = 27
        MULTI_CHANNEL_PULSED_SWEEP = 28


class MSC:
    class Post(IntEnum):
        BASE_VALUE = 1
        BIAS_VALUE = 2


class ODSW:
    class SwitchNormalState(IntEnum):
        NORMALLY_OPEN = 0
        NORMALLY_CLOSED = 1


class OSX:
    class Level(IntEnum):
        LOW = 0
        HIGH = 1
        EDGE = 2


class QSM:
    class Post(IntEnum):
        START = 1
        STOP = 2


class QSZ:
    class Mode(IntEnum):
        DISABLE = 0
        ENABLE = 1
        PERFORM_MEASUREMENT = 2


class RM:
    class Mode(IntEnum):
        DEFAULT = 1
        AUTO_UP = 2
        AUTO_UP_DOWN = 3


class SAP:
    class Path(IntEnum):
        SMU = 0
        AUX = 1


class SIM:
    class Mode(IntEnum):
        PULSE_GEN = 0
        ARB_WAVE_GEN = 1


class SPM:
    class Mode(IntEnum):
        DC_VOLTAGE = 0
        TWO_LEVEL_PULSE_SOURCE_1 = 1
        TWO_LEVEL_PULSE_SOURCE_2 = 2
        THREE_LEVEL_PULSE_SOURCE_1_AND_2 = 3


class SPRM:
    class Mode(IntEnum):
        FREE_RUN = 0
        COUNT = 1
        DURATION = 2


class SPT:
    class Src(IntEnum):
        PULSE_SRC_1 = 1
        PULSE_SRC_2 = 2


class SPV:
    class Src(IntEnum):
        DC_BIAS_SRC = 0
        PULSE_SRC_1 = 1
        PULSE_SRC_2 = 2


class SRE(IntFlag):
    DATA_READY = 1
    WAIT = 2
    INTERLOCK_OPEN = 8
    SET_READY = 16
    ERROR = 32
    RQS = 64


class SSP:
    class Path(IntEnum):
        FORCE1SENSE1_OPEN = 1
        OPEN_FORCE2SENSE2 = 2
        FORCE1SENSE1_FORCE2SENSE2 = 3
        CMUH_CMUL = 4


class STGP:
    class TriggerTiming(IntEnum):
        DISABLE_TRIGGER = 0
        SYNC_TO_PG_PULSE_OR_START_OF_ALWG_SEQUENCE = 1
        ALWG_PATTERN_CHANGE_OR_START_OF_FIRST_PATTERN = 2
        START_OF_EVERY_ALWG_PATTERN = 3


class TGMO:
    class Mode(IntEnum):
        EDGE = 1
        GATE = 2


class TGP:
    class TerminalType(IntEnum):
        INPUT = 1
        OUTPUT = 2

    class Polarity(IntEnum):
        POSITIVE = 1
        NEGATIVE = 2

    class TriggerType(IntEnum):
        """
        The meaning of the Trigger Type value depends on the value of
        TerminalType
        """
        ZERO = 0
        ONE = 1
        TWO = 2
        THREE = 3


class TGSI:
    class Mode(IntEnum):
        CASE1 = 1
        CASE2 = 2


class TGSO:
    class Mode(IntEnum):
        EDGE = 1
        GATE = 2


class TGXO:
    class Mode(IntEnum):
        EDGE = 1
        GATE = 2


class TM:
    class Mode(IntEnum):
        XE_CMD_AND_GPIB_GET = 1
        XE_CMD = 2
        XE_CMD_AND_EXT_TRIGGER = 3
        XE_CMD_AND_MM_CMD = 4


class TMACV:
    class Range(StrEnum):
        MAX_0V016 = '0.016'
        MAX_0V032 = '0.032'
        MAX_0V064 = '0.064'
        MAX_0V125 = '0.125'
        MAX_0V250 = '0.250'


class TMDCV:
    class Range(IntEnum):
        MFCMU_8V = 8
        MFCMU_12V = 12
        MFCMU_25V = 25
        SMU_100V = 100


class TST:
    class Option(IntEnum):
        RETURN_PASS_FAIL = 0
        PERFORM_TEST_AND_RETURN_PASS_FAIL = 1


class TTC:
    class Range(IntEnum):
        MAX_100OHM = 50
        MAX_300OHM = 100
        MAX_1KOHM = 300
        MAX_3KOHM = 1000
        MAX_10KOHM = 3000
        MAX_30KOHM = 10000
        MAX_100KOHM = 30000
        MAX_300KOHM = 100000
        MAX = 300000


class UNT:
    class Mode(IntEnum):
        MODULE_INFO_ONLY = 0
        MODULE_AND_MAINFRAME_INFO = 1


class VAR:
    class Type(IntEnum):
        INTEGER = 0
        FLOAT = 1


class WAT:
    class Type(IntEnum):
        SMU_SOURCE_WAIT_TIME = 1
        SMU_MEASURE_WAIT_TIME = 2
        MFCMU_MEASUREMENT_WAIT_TIME = 3


class WM:
    class Post(IntEnum):
        START = 1
        STOP = 2


class WMACV:
    class Post(IntEnum):
        START = 1
        STOP = 2


class WMDCV:
    class Post(IntEnum):
        START = 1
        STOP = 2


class WMFC:
    class Post(IntEnum):
        START = 1
        STOP = 2


class WNX:
    class Mode(IntEnum):
        VOLTAGE = 1
        CURRENT = 2


class WS:
    class Mode(IntEnum):
        CONTINUE_IMMEDIATELY_IF_PENDING_TRIGGER = 1
        WAIT_IMMEDIATELY = 2


class WSX:
    class Mode(IntEnum):
        CONTINUE_IMMEDIATELY_IF_PENDING_TRIGGER = 1
        WAIT_IMMEDIATELY = 2



"""
Test the auto-loadable channels and channels list. These channels are helpful
when dealing with channel types which can be added or deleted from an
instrument. Please note that `channel` in this context does not necessarily
mean a physical instrument channel, but rather an instrument sub-module.
"""

import re
from typing import Any, Callable, Dict, List, Optional, Union

import pytest

from qcodes import Instrument
from qcodes.instrument.channel import (
    AutoLoadableChannelList,
    AutoLoadableInstrumentChannel,
    InstrumentChannel,
)


class MockBackendBase:
    """
    A very simple mock backend that contains a dictionary with string keys and
    callable values. The keys are matched to input commands with regular
    expressions and on match the corresponding callable is called.
    """
    def __init__(self) -> None:
        self._command_dict: Dict[str, Callable[..., Any]] = {}

    def send(self, cmd: str) -> Any:
        """
        Instead of sending a string to the Visa backend, we use this function
        as a mock
        """
        keys = self._command_dict.keys()
        ans = None
        key = ""
        for key in keys:
            ans = re.match(key, cmd)
            if ans is not None:
                break

        if ans is None:
            raise ValueError(f"Command {cmd} unknown")

        args = ans.groups()
        return self._command_dict[key](*args)


class MockBackend(MockBackendBase):
    """
    A mock backend for our test instrument. It defines the following SCPI (like)
    commands:

    INST:CHN<n>:HLO
    Return the string "Hello from channel <n>" where n is a channel number

    INST:CHN:ADD <n>, <greeting>
    Add a channel with channel number n. We use the greeting when the hello
    parameter is called.

    INST:CHN:DEL <n>
    Delete a channel with channel number n

    INST:CHN:CAT
    Return a catalog of currently defined channels

    INST:CHN<n>:GRT
    Return the greeting of this channel
    """
    def __init__(self) -> None:
        super().__init__()
        self._channel_catalog: List[str] = ["1", "2", "4", "5"]  # Pre-existing
        # channels
        self._greetings = {chn: "Hello" for chn in self._channel_catalog}

        self._command_dict = {
            r":INST:CHN(\d):HLO":
                lambda chn: self._greetings[chn] + " from channel " + str(chn),
            r":INST:CHN:ADD (\d), (.+)": self._add_channel,
            r":INST:CHN:DEL (\d)": self._channel_catalog.remove,
            r":INST:CHN:CAT": lambda: ",".join(str(i) for i in self._channel_catalog),
            r":INST:CHN(\d):GRT": self._greetings.get,
        }

    def _add_channel(self, chn: int, greeting: str)->None:
        """
        Add a channel on the mock instrument
        """
        self._channel_catalog.append(str(chn))
        self._greetings[str(chn)] = greeting


class SimpleTestChannel(AutoLoadableInstrumentChannel):
    """
    A channel to test if we can create and delete channel instances
    """

    @classmethod
    def _discover_from_instrument(
            cls, parent: Instrument, **kwargs) -> List[Dict[Any, Any]]:
        """
        New channels need `name` and `channel` keyword arguments.
        """
        channels_str = parent.channel_catalog()
        channels_to_skip = kwargs.get("channels_to_skip", [])  # Note that
        # `channels_to_skip` is an optional kwarg for loading from instrument.
        # We test this by giving this keyword during the initialization of the
        # AutoLoadableChannelList.
        kwarg_list = []
        for channel_str in channels_str.split(","):

            if channel_str in channels_to_skip:
                continue

            channel = int(channel_str)
            greeting = parent.ask(f":INST:CHN{channel}:GRT")
            new_kwargs = {
                "name": f"channel{channel}",
                "channel": channel,
                "greeting": greeting
            }
            kwarg_list.append(new_kwargs)

        return kwarg_list

    @classmethod
    def _get_new_instance_kwargs(
            cls, parent: Optional[Instrument] = None, **kwargs
    ) -> Dict[Any, Any]:
        """
        Find the smallest channel number not yet occupied. An optional keyword
        `greeting` is extracted from the kwargs. The default is "Hello"
        """
        if parent is None:
            raise RuntimeError("SimpleTestChannel needs a parent instrument")
        channels_str = parent.channel_catalog()
        existing_channels = [int(i) for i in channels_str.split(",")]

        new_channel = 1
        while new_channel in existing_channels:
            new_channel += 1

        new_kwargs = {
            "name": f"channel{new_channel}",
            "channel": new_channel,
            "greeting": kwargs.get("greeting", "Hello")
        }

        return new_kwargs

    def __init__(
            self,
            parent: Union[Instrument, InstrumentChannel],
            name: str,
            channel: int,
            greeting: str,
            existence: bool = False,
            channel_list: Optional[AutoLoadableChannelList] = None,
    ) -> None:

        super().__init__(parent, name, existence, channel_list)
        self._channel = channel
        self._greeting = greeting

        self.add_parameter(
            "hello",
            get_cmd=f":INST:CHN{self._channel}:HLO"
        )

    def _create(self) -> None:
        """Create the channel on the instrument"""
        self.parent.root_instrument.write(
            f":INST:CHN:ADD {self._channel}, {self._greeting}")

    def _remove(self) -> None:
        """Remove the channel from the instrument"""
        self.write(f":INST:CHN:DEL {self._channel}")


class DummyInstrument(Instrument):
    """
    This dummy instrument allows the creation and deletion of
    channels
    """

    def __init__(self, name: str)->None:
        super().__init__(name)

        self._backend = MockBackend()

        self.add_parameter(
            "channel_catalog",
            get_cmd=":INST:CHN:CAT",
        )

        channels = AutoLoadableChannelList(
            self, "channels", SimpleTestChannel, channels_to_skip=["5"])
        self.add_submodule("channels", channels)

    def write_raw(self, cmd: str) -> None:
        self._backend.send(cmd)

    def ask_raw(self, cmd: str) -> str:
        return self._backend.send(cmd)


@pytest.fixture(scope='function')
def dummy_instrument():
    instrument = DummyInstrument("instrument")
    yield instrument
    instrument.close()


def test_sanity(dummy_instrument):
    """
    Test the basic functionality of the auto-loadable channels, without using
    the auto-loadable channels list. Please note that the `channels_to_skip`
    argument in the dummy instrument applies when accessing channels
    from the channels list. Since we are calling `load_from_instrument` directly
    in this test without this keyword argument, we will see all channels.
    """
    # Test that we are able to discover instruments automatically
    channels = SimpleTestChannel.load_from_instrument(dummy_instrument)
    assert len(channels) == 4
    assert channels[0].hello() == "Hello from channel 1"
    assert channels[1].hello() == "Hello from channel 2"
    assert channels[2].hello() == "Hello from channel 4"
    assert channels[3].hello() == "Hello from channel 5"
    # Test that we can generate a new instance of the channels without
    # conflicting names
    new_channel_kwargs = SimpleTestChannel._get_new_instance_kwargs(
        dummy_instrument)

    new_channel = SimpleTestChannel(dummy_instrument, **new_channel_kwargs)
    # Instrument IO through the newly instantiated channel should raise an
    # exception before actually creating the channel on the instrument
    with pytest.raises(
            RuntimeError,
            match=r"Object does not exist \(anymore\) on the instrument"):

        new_channel.hello()

    # After creating the channel we should be able to talk to the instrument.
    new_channel.create()
    assert new_channel.hello() == "Hello from channel 3"

    # Once we remove the channel we should no longer be able to talk to the
    # instrument
    new_channel.remove()
    with pytest.raises(
            RuntimeError,
            match=r"Object does not exist \(anymore\) on the instrument"):

        new_channel.hello()  # We have deleted the channel and it should no
        # longer be available


def test_channels_list(dummy_instrument):
    """
    Test the auto-loadable channels list
    """
    assert dummy_instrument.channels[0].hello() == "Hello from channel 1"
    assert dummy_instrument.channels[1].hello() == "Hello from channel 2"
    assert dummy_instrument.channels[2].hello() == "Hello from channel 4"
    assert len(dummy_instrument.channels) == 3
    # Test that we can add channels
    new_channel = dummy_instrument.channels.add()
    assert len(dummy_instrument.channels) == 4
    assert new_channel is dummy_instrument.channels[-1]
    assert new_channel.hello() == "Hello from channel 3"
    # Test that we can remove them
    new_channel.remove()
    assert len(dummy_instrument.channels) == 3
    assert new_channel not in dummy_instrument.channels
    # Once removed we should no longer be able to talk to the channel
    with pytest.raises(
            RuntimeError,
            match=r"Object does not exist \(anymore\) on the instrument"):

        new_channel.hello()
    # Remove a channel that was pre-existing on the instrument.
    dummy_instrument.channels[-1].remove()
    assert len(dummy_instrument.channels) == 2


def test_with_kwargs(dummy_instrument):
    """
    Test keyword arguments given to the add method
    """
    new_channel = dummy_instrument.channels.add(greeting="Guden tag")
    assert new_channel.hello() == "Guden tag from channel 3"
